# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import re
import os
import shutil
import time
import datetime
# import tracemalloc    # noqa
# import gc
import pathlib
import sys
import stat

# https://mariadb.com/docs/server/connect/programming-languages/c/install/
# noqa      sudo apt install libmariadb3 libmariadb-dev
# pip install mariadb

try:
    import mariadb
except ImportError as err:
    print(f"Error {err} while importing mariadb")
    mariadb = None

import hashlib

# pip install fabric
# doc: https://docs.fabfile.org/en/stable/
# doc: https://www.paramiko.org/
# doc: https://help.ubuntu.com/community/SSH/OpenSSH/Keys
import fabric
import invoke

# pip install requests
# doc: https://requests.readthedocs.io/en/latest/

# import requests.cookies


import urllib3.exceptions
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
# Remove useless warning on https certificates

from piwiPre.pwpErrors import LOGGER, PwpError, PwpConfigError, PwpInternalError


class PwpSummary:
    def __init__(self, stage, old_file):
        self.stage = stage
        self.old_file = old_file
        self.local = old_file.is_local
        self.remote = old_file.is_remote
        self.author = ''
        self.date = ''
        self.rotation = ''

        self.meta_date = False
        self.meta_instructions = False
        self.meta_author = False
        self.meta_copyright = False

        self.action = ""
        self.destination = None

        self.backup = ""

        self.representative = False

        self.db_created = False
        self.db_size = False
        self.db_width = False
        self.db_height = False
        self.db_md5 = False
        self.db_gps = False
        self.db_author = False

        self.thumb_sq = False
        self.thumb_th = False
        self.thumb_me = False
        self.thumb_cu = False
        self.thumb_index = False

        self.rem_thumb_sq = False
        self.rem_thumb_th = False
        self.rem_thumb_me = False
        self.rem_thumb_cu = False
        self.rem_thumb_index = False

        self.auto_conf = False
        self.rem_auto_conf = False

    def get_meta(self):
        meta = ""
        meta += "d" if self.meta_date else '-'
        meta += "i" if self.meta_instructions else '-'
        meta += "a" if self.meta_author else '-'
        meta += "c" if self.meta_copyright else '-'
        return meta

    def __str__(self):
        name = f"{self.old_file.basename[-40:]:40}"
        author = f"{self.author[-15:]:15}"

        action = self.action
        dst = '' if self.destination is None else self.destination.local
        # dst = f"{destination[-50:]:50}"

        backup = f"{self.backup}"

        where = ""
        where += 'L' if self.local else '-'
        where += 'R' if self.remote else '-'

        meta = self.get_meta()

        rep = "R" if self.representative else '-'

        db = ""
        db += "c" if self.db_created else '-'
        db += "s" if self.db_size else '-'
        db += "w" if self.db_width else '-'
        db += "h" if self.db_height else '-'
        db += "5" if self.db_md5 else '-'
        db += "g" if self.db_gps else '-'
        db += "a" if self.db_author else '-'

        thumbs = ""
        thumbs += 'S' if self.thumb_sq else '-'
        thumbs += 'T' if self.thumb_th else '-'
        thumbs += 'M' if self.thumb_me else '-'
        thumbs += 'C' if self.thumb_cu else '-'
        thumbs += 'I' if self.thumb_index else '-'

        rem_thumbs = ""
        rem_thumbs += 's' if self.rem_thumb_sq else '-'
        rem_thumbs += 't' if self.rem_thumb_th else '-'
        rem_thumbs += 'm' if self.rem_thumb_me else '-'
        rem_thumbs += 'c' if self.rem_thumb_cu else '-'
        rem_thumbs += 'i' if self.rem_thumb_index else '-'

        return (f"{self.stage:6} {name} LR[{where}] A[{author}] D[{self.date:19}] " +
                f"rot[{self.rotation:2}] meta[{meta}] rep[{rep}] db[{db}] " +
                f"th[{thumbs}] rth[{rem_thumbs}] {action:5}:[{dst}] back[{backup}]")


class DirInfo:
    def __init__(self, name, path, dir_id, rank, id_upper_cat, global_rank, upper_cats):
        self.name = name
        self.path = path        # full path starting with ./galleries/photo  (not available in the database)
        self.dir_id = dir_id
        self.rank = rank
        self.id_upper_cat = id_upper_cat
        self.global_rank = global_rank
        self.upper_cats = upper_cats

    def __str__(self):
        return f"dir('{self.path}':{self.dir_id})"

    #    images: piwigo_images                                      updated if file changed     created if file new
    #       id: unique id of the picture                                                                yes
    #       file: filename, without path                                                                yes
    #       date_available: date on insertion in the database                                           yes
    #       date_creation: creation date of the file                                                    yes
    #       name: display name of the picture. defaults to file                                         yes
    #       comment: user comment. defaults to null
    #       author: defaults to null
    #       hit:
    #       filesize: file size                                             yes,                        yes     # noqa
    #       width:  picture width                                           yes,                        yes
    #       height: picture height                                          yes,                        yes
    #       coi:
    #       representative_ext: the 4 char extension of representative
    #       date_metadata_update:                                                                       yes
    #       path:  full path of file, starting with ./galleries/photo                                   yes
    #       storage_category_id: id of storage dir                                                      yes
    #       level:  level of privacy. defaults to 0. sometimes 4, which is probably an error            yes
    #       md5sum:                                                         yes                         yes
    #       added_by: id of inserter,                                                                   yes
    #       rotation: defaults to 0                                         yes,                        yes
    #       latitude:                                                       yes,                        yes
    #       longitude:                                                      yes,                        yes
    #               see 2023-02-19-11h02-55-Plouhinec.jpg                                                       # noqa
    #       lastmodified:                                                   yes (automatic)             yes     # noqa


class FileInfo:
    def __init__(self, file_id, file, date_available, date_creation, name, author, file_size, width, height,
                 date_metadata_update,  path, storage_category_id, level, md5sum, added_by, latitude, longitude,
                 representative_ext, last_modified):
        self.file_id = file_id
        self.file = file
        self.date_available = date_available
        self.date_creation = date_creation
        self.name = name
        self.author = author
        self.file_size = file_size
        self.width = width
        self.height = height
        self.date_metadata_update = date_metadata_update
        self.path = path
        self.storage_category_id = storage_category_id
        self.level = level
        self.md5sum = md5sum if md5sum != "null" else None
        self.added_by = added_by
        self.latitude = latitude
        self.longitude = longitude
        self.representative_ext = representative_ext
        self.last_modified = last_modified

    def __str__(self):
        return f"file('{self.path}':{self.file_id})"


class PwpActor:
    allowed_chars = r"a-zA-Z0-9\-_.&@~!,;+°()àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ "  # noqa
    tmp_dir = '.piwiPre.tmp'

    def __init__(self):
        self.dryrun = False
        # values that are cached from the 1st config used
        self.print_debug = False
        self.trace_malloc = False

        # management of piwigo
        self.piwigo_user = None
        self.piwigo_level = 0

        # management of ssh/sftp
        self.remote_user = None
        self.remote_pwd = None
        self.remote_host = None
        self.remote_port = None
        self.ls_command = None
        self.ls_output = None
        self.ssh_connection = None

        self.remote_uname = None

        self.sql_host = None
        self.sql_port = None
        self.sql_user = None
        self.sql_pwd = None
        self.sql_database = None
        self.sql_connection = None

        # info about the first album
        self.sql_first_album = None     # name
        self.sql_first_album_id = None  # id of the first album
        self.sql_first_album_path = None
        self.sql_first_album_upper_cat = None
        self.sql_first_album_global_rank = None
        self.sql_first_album_upper_cats = None
        self.sql_first_album_rank = None

        # end of cache
        self.dir_made = []
        self.dir_numbers = {}

    def configure(self, config):
        self.print_debug = config['debug']
        self.dryrun = config['dryrun']
        self.trace_malloc = config['trace-malloc']

        #  self.trace_malloc_start()

        self.connect_ssh(config)
        self.connect_sql(config)

    def reset_data(self):
        self.dryrun = False
        # values that are cached from the 1st config used
        self.print_debug = False
        # end of cache
        self.dir_made = []
        self.dir_numbers = {}

        # self.ssh_connection = None  NO REASON to do this, notably without closing the previous one

    @staticmethod
    def isdir(path):
        # os.path.isdir() is NOT reliable on NFS share drives
        try:
            res = os.stat(path)
            return stat.S_ISDIR(res.st_mode)
        except FileNotFoundError:
            # this is the normal error
            return False
        except NotADirectoryError:
            # this is another normal error
            return False
        except OSError as e:
            if e.errno == 6:
                # non valid descriptor
                return False
            raise PwpError(f"Error {e}, errno = {e.errno} while isdir({path})")  # noqa

    @staticmethod
    def isfile(path):
        return os.path.isfile(path)

    def mkdirs(self, dir_name, forced=False):
        dir_name = dir_name.rstrip('/')
        dir_name = dir_name or '.'

        if self.dryrun and not forced:
            LOGGER.debug(f"Would makedirs '{dir_name}'")
            return True

        # # ACTOR.flush_nfs_dir(os.path.dirname(dir_name))     # noqa
        # the father dir may not exist, so we cannot flush it

        father = os.path.dirname(dir_name)
        if os.path.dirname(father) != father:
            # os.path.dirname('//toto/fifi/lulu") -> '//toto/fifi'  # noqa
            # BUT:
            # os.path.dirname('//toto/fifi") -> '//toto/fifi', because python recognizes the network share syntax/ # noqa
            self.mkdirs(father)
            # this is a home-made recursive implementation of makedirs,
            # in an attempt to trap all nfs flush problems.

        # if father != dir_name:
            # ACTOR.flush_nfs_dir(father)

        if self.isdir(dir_name):
            # LOGGER.msg(f"mkdirs: '{dir_name}' exists")
            # ACTOR.flush_nfs_dir(dir_name)
            return False

        try:
            os.makedirs(dir_name, exist_ok=True)

        except FileExistsError as e:
            # this should NEVER happen, because we have just checked if os.path.isdir(dir_name) ...
            # but, in real-life, it DOES happen, probably some inconsistency in the NFS implementation
            raise PwpError(f"mkdirs: '{dir_name}' exists ERROR {e}")
        except OSError as e:
            raise PwpError(f"mkdirs: '{dir_name}' ERROR {e}")

        # ACTOR.flush_nfs_dir(father)

        if self.isdir(dir_name):
            LOGGER.debug(f"makedirs '{dir_name}'")
        else:
            # This is probably a transient error due to NFS. Better stop and rerun than keep an inconsistent state
            raise PwpError(f"FAIL: makedirs '{dir_name}'")

    def copy(self, src, dst, forced=False):
        """
        copy src to dst, unless dryrun is True
        :param src: file to copy
        :param dst: destination filename
        :param forced: if True, copy is always done, if False, do not copy if dryrun is True
        :return: None
        """
        base = os.path.dirname(dst)
        self.mkdirs(base, forced)

        if not ACTOR.isfile(src):
            raise PwpError(f"FAILED copy '{src}' ->  '{dst}' : non existing source")

        if self.dryrun and not forced:
            LOGGER.debug(f"Would copy '{src}' ->  '{dst}'")
            return

        shutil.copy2(src, dst)  # preserve metadata

        # ACTOR.flush_nfs_dir(os.path.dirname(dst))     # noqa

        if ACTOR.isfile(dst):
            LOGGER.debug(f"copy '{src}' ->  '{dst}'")
        else:
            # This is probably a transient error due to NFS. Better stop and rerun than keep an inconsistent state
            raise PwpError(f"FAIL:copy '{src}' ->  '{dst}'")

    def backup(self, filename):
        if not ACTOR.isfile(filename):
            return

        m1 = re.match(r"(.*)\.bak$", filename)
        m2 = re.match(r"(.*)\.bak-(\d*)$", filename)
        if m1:
            bak = f"{m1.group(1)}.bak-1"
        elif m2:
            num = int(m2.group(2))
            bak = f"{m2.group(1)}.bak-{num + 1}"
        else:
            bak = filename + '.bak'

        self.backup(bak)
        self.copy(filename, bak)

    def copytree(self, src, dst):
        """
        copytree(self, src, dst): safely copy src to dst : dst will be a copy of src
        :param src: source directory
        :param dst: destination directory
        :return: None
        """
        if self.dryrun:
            LOGGER.debug(f"Would copytree'{src}' ->  '{dst}'")
            return

        shutil.copytree(src, dst, dirs_exist_ok=True)

        # ACTOR.flush_nfs_dir(dst)

        if ACTOR.isdir(dst):
            LOGGER.debug(f"copytree '{src}' ->  '{dst}'")
        else:
            # This is probably a transient error due to NFS. Better stop and rerun than keep an inconsistent state
            raise PwpError(f"FAIL:copytree '{src}' ->  '{dst}'")

    def move(self, src, dst, forced=False):
        if self.dryrun and not forced:
            LOGGER.debug(f"Would move file '{src}' -> '{dst}'")
            return

        base = os.path.dirname(dst)
        self.mkdirs(base)

        if os.path.exists(dst):
            self.delete(dst)  # pragma: no cover

        shutil.move(src, dst)

        # ACTOR.flush_nfs_dir(os.path.dirname(dst)) # noqa

        # let's assume the library does the job and yell in case of problem
        #
        if os.path.exists(dst):
            LOGGER.debug(f"move file '{src}' -> '{dst}'")
        else:
            # from time to time, this exception arises which is completely abnormal,
            # since shutil.move has raised no exception
            # this happens ONLY on NFS network shares, not with physical HDD
            raise PwpError(f"FAIL:move file '{src}' -> '{dst}'")

    def delete(self, src, forced=False, msg=""):
        # if we do not do this, we have file not found exceptions
        # because in some tests, the file has just been created, erased, created again
        # and NFS buffers are outdated
        # ACTOR.flush_nfs_dir(os.path.dirname(src)) # noqa

        if not os.path.exists(src):
            LOGGER.debug(f"CAVEAT: {msg} Delete '{src}' non existing file")
            return False

        if self.dryrun and not forced:
            LOGGER.info(f"Would delete '{src}' {msg}")
            return False

        try:
            os.remove(src)
        except FileNotFoundError as e:
            # this should NEVER happen, but does from time to time,
            # and always end up with the file being erased
            LOGGER.msg(f"delete file  '{src}' raised FileNotFoundError Exception {e}")
            raise PwpError(f"FAIL: delete '{src}'")
        except OSError as e:
            raise PwpError(f"FAIL: delete '{src}' error {e}")

        # ACTOR.flush_nfs_dir(os.path.dirname(src)) # noqa

        if os.path.exists(src):
            raise PwpError(f"FAIL: delete '{src}'  {msg}")

        LOGGER.debug(f"deleted file '{src}'  {msg} OK")
        return True

    def rmtree(self, src, msg: str = ''):
        if not ACTOR.isdir(src):
            LOGGER.debug(f"rmtree '{src}' : non existing directory  {msg}")
            return False

        if self.dryrun:
            LOGGER.debug(f"would remove tree '{src}'  {msg}")
            return False

        try:
            shutil.rmtree(src)  # self.do_rmtree(src)  #
        except OSError as e:
            raise PwpError(f"FAIL: error {e} while remove tree '{src}'  {msg}")

        # ACTOR.flush_nfs_dir(os.path.dirname(src)) # noqa

        if ACTOR.isdir(src):
            # This is probably a transient error due to NFS. Better stop and rerun than keep an inconsistent state
            raise PwpError(f"FAIL:  remove tree '{src}'  {msg}")
        else:
            LOGGER.debug(f"removed tree  '{src}'  {msg}")

        # Python bug here !
        #       We have from time to time, located on the calling context in PwpMain:
        #       UnboundLocalError: local variable 'local_file_path' referenced before assignment
        #       when this is error is clearly wrong.
        #       It seems that adding a return value helps.
        return True

    @staticmethod
    def open(filename: str, mode: str):
        # if self.dryrun:
        #     LOGGER.debug(f"open '{filename}' ")
        #     return

        if mode == 'r' and not ACTOR.isfile(filename):
            raise PwpError(f"ERROR reading non-existing file {filename}")
        if "b" in mode:
            return open(filename, mode)  # binary mode doesn't take an encoding argument
        return open(filename, mode, encoding="utf-8")

    @staticmethod
    def get_last_numerical_field(template: str):
        items_list = re.findall('{[^}]*}', template)
        non_numerical = ['{a', '{month_name}', '{base}', '{author}', '{suffix}', '{file}']
        while items_list:
            res = items_list.pop()
            if res not in non_numerical:
                return res[1:-1]  # skip {}
        raise PwpInternalError(f"incoherent get_last_numerical_field {template} returns None")

    def get_info_from_format(self, template: str, src: str):
        """
        Extract information from src according to the descriptor
        :param template: a string that describes the information format
        :param src: the source with information
        :return: a dictionary

        Assuming that template is a reasonably simple format,
        notably that the same field does not occur 2 times.

        Assuming also that the possible items within template are all known,
        which is the case in piwiPre.

        If src is the result of formatting template with some values,
        then we can find back the values, provided the string is simple enough.
        This can even been done independently of the order of the fields in the template,
        because we can find the order by analysing template.
        """
        items_list = re.findall('{[^}]*}', template)  # the list of all items to find, e.g. '{Y}'
        # here, we have all items possibly found in piwiPre
        trans = {
            'size': r"(\d+)",      # noqa
            'Y': r"(\d\d\d\d)",
            'm': r"(\d\d)",
            'd': r"(\d\d)",
            'H': r"(\d\d)",
            'M': r"(\d\d)",
            'S': r"(\d\d)",
            'ms': r"(\d+)",
            'count': r"(\d+)",
            'z': r"(\+?\-?\d\d\d\d)",
            'a': r"(am|pm)",
            'month_name': '([' + self.allowed_chars + ']+)',
            'base': '([' + self.allowed_chars + ']+)',
            'author': r'(.*)',
            'suffix': r'(\w+)$',
            'file': r"(.*?)",  # noqa
            'flags': r'([-+\w]+)',
        }
        str_format = template.format(**trans)
        res = re.match(str_format, src)
        dico = None
        if res:
            dico = {}
            for field in trans.keys():
                ff = '{'+field+'}'
                dico[field] = res.group(items_list.index(ff) + 1) if ff in items_list else None
        return dico

    def create(self, filename):
        with self.open(filename, 'w') as f:
            f.write(f"Fake file created for test {datetime.datetime.now()}\n")

    # ----------------------------------------------------------------------
    # management of ssh/sftp

    @staticmethod
    def build_timestamp(filename: str):
        file_time = os.path.getmtime(filename)
        timestamp = time.strftime("%Y/%m/%d-%H:%M:%S", time.localtime(file_time))
        return timestamp

    @staticmethod
    def timestamp_from_ls(d: dict):
        dt: datetime.datetime = d['date']
        timestamp = f"{dt.year:4}/{dt.month:02}/{dt.day:02}-{dt.hour:02}:{dt.minute:02}:{dt.second:02}"
        return timestamp

    def remote_run(self, cmd: str, forced=False, warn_if_error=True):
        if self.dryrun and not forced:
            return None

        if self.ssh_connection:
            try:
                res = self.ssh_connection.run(cmd, hide=True, warn=True, encoding='utf8')
            except invoke.exceptions.ThreadException as e:
                LOGGER.warning(f"ssh('{cmd}') failed with unexpected exception {e}")
                return None

            if not res.ok:
                LOGGER.info(f"CAVEAT remote '{cmd}' returned {res.stderr}")
                if warn_if_error:
                    LOGGER.warning(f"CAVEAT remote '{cmd}' returned {res.stderr}")

            return res
        raise PwpError("trying to run a ssh command without a ssh connection", cmd)

    @staticmethod
    def my_decode(item: str) -> str:
        """
        Manages the unknown encoding used by ls - paramiko - fabfile  
        returns the decoded string
        only chars in 'allowed_chars' are processed
        
        :param item: string to be decoded 
        :return: the decoded string"""    # noqa

        # only allowed chars
        table = {'\\302\\260': '°', '\\303\\240': 'à', '\\303\\242': 'â', '\\303\\244': 'ä', '\\303\\251': 'é',
                 '\\303\\250': 'è', '\\303\\252': 'ê', '\\303\\253': 'ë', '\\303\\257': 'ï', '\\303\\256': 'î',
                 '\\303\\264': 'ô', '\\303\\266': 'ö', '\\303\\271': 'ù', '\\303\\273': 'û', '\\303\\274': 'ü',
                 '\\303\\277': 'ÿ', '\\303\\247': 'ç', '\\303\\261': 'ñ', '\\303\\200': 'À', '\\303\\202': 'Â',
                 '\\303\\204': 'Ä', '\\303\\211': 'É', '\\303\\210': 'È', '\\303\\212': 'Ê', '\\303\\213': 'Ë',
                 '\\303\\217': 'Ï', '\\303\\216': 'Î', '\\303\\224': 'Ô', '\\303\\226': 'Ö', '\\303\\231': 'Ù',
                 '\\303\\233': 'Û', '\\303\\234': 'Ü', '\\305\\270': 'Ÿ', '\\303\\207': 'Ç', '\\303\\221': 'Ñ'}
        new_val = ''
        i = 0
        while i < len(item):
            s = item[i:i+8]
            if s in table:
                new_val += table[s]
                i += 8
            elif item[i:i+2] == '\\\\':
                new_val += '\\'
                i += 2
            else:
                new_val += item[i]
                i += 1
        return new_val

    def remote_ls(self, directory, forced=False, warn_if_absent=False):
        directory = directory or '.'
        LOGGER.debug(f"ssh ls '{directory}' ")
        if not forced and self.dryrun:
            return {}  # pragma: no cover

        ls_cmd = self.ls_command.format(file=directory)
        try:
            result = self.remote_run(ls_cmd, forced=forced, warn_if_error=warn_if_absent)
        except FileNotFoundError:
            return {}
        res = self.my_decode(result.stdout)
        all_lines = res.split('\n')
        all_files: dict = {}
        for line in all_lines:
            dico = self.get_info_from_format(self.ls_output, line)
            if dico:
                f_date = datetime.datetime(year=int(dico['Y']), month=int(dico['m']), day=int(dico['d']),
                                           hour=int(dico['H']), minute=int(dico['M']), second=int(dico['S']),
                                           microsecond=int(int(dico['ms'])/1000) if 'ms' in dico else 0)
                # NB: datetime gets microseconds, but ls provides nanoseconds
                # TODO: manage timezone
                if dico['flags'][0] == 'd':
                    new_dico = {"date": f_date, "dir_name": dico['file'], "type": 'dir'}
                else:
                    new_dico = {"date": f_date, "size": int(dico['size']), "filename": dico['file'], 'type': 'file'}
                all_files[dico["file"]] = new_dico

        return all_files

    def remote_create(self, filename):
        self.remote_run(f"touch {filename}")

    def remote_isfile(self, filepath, forced=False, warn_if_absent=False):
        """

        :param filepath: file path , on the remote host, of file to test
        :param forced: if True, do it even if dryrun is True
        :param warn_if_absent: if True, issues a warning if file is absent
        :return: None if the file does not exist, dico of information if file exists
        """
        if self.dryrun and not forced:
            return None  # pragma: no cover: defensive code
        all_files = self.remote_ls(os.path.dirname(filepath), forced=forced, warn_if_absent=warn_if_absent)
        if all_files is not None and os.path.basename(filepath) in all_files.keys():
            LOGGER.debug(f"ssh file_exists '{filepath}' : YES")
            return all_files[os.path.basename(filepath)]
        LOGGER.debug(f"ssh file_exists '{filepath}' : NO")
        return None

    def remote_mkdir(self, directory):
        if directory in self.dir_made:
            return
        LOGGER.debug(f"remote mkdir '{directory}'")
        self.remote_run(f'mkdir -p "{directory}"')  # need "" to quote spaces
        self.dir_made.append(directory)                 # self.dir_made is actually be cleaned at the end of run

    def remote_put(self, src, directory):
        LOGGER.debug(f"remote put '{src}' '{directory}'")
        if self.dryrun:
            return
        tmp_file = self.tmp_dir + '/' + src
        if self.ssh_connection:
            tmp_path = os.path.dirname(tmp_file)
            self.remote_mkdir(tmp_path)
            self.remote_mkdir(directory)
            sftp = self.ssh_connection.sftp()
            sftp.put(src, tmp_file, confirm=True)
            f_a_time = os.path.getatime(src)
            f_m_time = os.path.getmtime(src)
            sftp.utime(tmp_file, (f_a_time, f_m_time))

        self.remote_run(f'mv -vf "{tmp_file}" "{directory}"')

    def remote_get(self, remote_file, local_file):
        LOGGER.debug(f"remote get '{remote_file}' -> '{local_file}'")
        # assuming  directory for local exists
        if self.dryrun:
            return
        if self.ssh_connection:
            self.remote_mkdir(self.tmp_dir)
            tmp_file = self.tmp_dir + '/' + os.path.basename(local_file)
            # -p: preserve date -v: verbose -f: force (clobbers)
            self.remote_run(f'cp -pvf {remote_file} {tmp_file}')      # noqa
            sftp = self.ssh_connection.sftp()
            local_dir = os.path.dirname(local_file)                   # noqa
            self.mkdirs(local_dir)
            sftp.get(tmp_file, local_file)
        else:
            raise PwpError("trying to run a ssh command without a ssh connection",
                           f"remote get '{remote_file}' -> '{local_file}'")

    def remote_compute_md5(self, remote_file):
        LOGGER.debug(f"remote compute_md5 '{remote_file}'")
        if self.dryrun:
            return
        result = self.remote_run(f"md5sum '{remote_file}'")
        if result is None:
            raise PwpInternalError(f"Unknown ssh error while remote compute_md5 '{remote_file}'")
        if not result.ok:
            raise PwpInternalError(f"ssh error {result.exited} while remote compute_md5 '{remote_file}'")

        # output of md5sum: 'aa8fe00349ca160e8bf0f88f45f5cea7  /volume1/ph ... -cup.jpg'
        res = result.stdout.split()[0]
        return res

    def remote_delete(self, filename: str, msg: str = ''):
        LOGGER.info(f"remote rm '{filename}'  {msg}")
        LOGGER.msg(f"remote delete '{filename}' {msg}")
        self.remote_run(f'rm -f "{filename}"')            # if no connection, falls into remote_run() error raise

    def remote_move(self, src: str, dst: str):
        LOGGER.debug(f"remote mv '{src}' -> '{dst}'")
        # assuming  directory for remote exists
        self.remote_run(f'mv -vf "{src}"  "{dst}"')     # if no connection, falls into remote_run() error raise

    # for now, unused
    # def remote_copy(self, src: str, dst: str):
    #     LOGGER.debug(f"remote mv '{src}' -> '{dst}'")
    #     # assuming  directory for remote exists
    #     self.remote_run(f'cp -vf {src}  {dst}')     # if no connection, falls into remote_run() error raise # noqa

    def remote_rmtree(self, src: str, msg: str = ''):
        LOGGER.info(f"remote rmdir '{src}' {msg}")
        # LOGGER.msg(f"remote rmdir '{src}' {msg}")
        self.remote_run(f'rm -rf "{src}"')            # if no connection, falls into remote_run() error raise

    def connect_ssh(self, config: dict):
        """
        :param config: configuration
        :return: Connection_done: bool, uname: str, cause:str, Error: bool
        """
        self.remote_host = config['remote-host']
        self.remote_port = int(config['remote-port']) if config['remote-port'] else None
        self.remote_user = config['remote-user']
        self.remote_pwd = config['remote-pwd']

        self.piwigo_user = config['piwigo-user']
        self.piwigo_level = config['piwigo-level']

        self.ls_output = config['ls-output']
        self.ls_command = config['ls-command']

        if self.ssh_connection is not None:
            return True, self.remote_uname, None, False

        # CAVEAT: here, we want to connect to SSH even if enable-remote-album or enable-remote-web are False,
        # because we will turn one of them True in the course of tests.

        remote = (config['remote-album'] or config['remote-web']) and self.remote_host and self.remote_port

        if not remote:
            return False, None, 'No remote parameters', False   # not connecting is not error here

        LOGGER.info(f"connect host='{self.remote_host}' port='{self.remote_port}' user='{self.remote_user}'")
        if self.dryrun:
            return False, None, "Dryrun", False  # not connecting is not error here

        self.ssh_connection = fabric.Connection(self.remote_host, self.remote_user, self.remote_port)
        if self.ssh_connection is None:
            LOGGER.info("SSH error while Connecting")
            return False, None, "SSH error Connecting", True

        self.ssh_connection.open()
        # by default, self.ssh_connection.open() returns None
        if not self.ssh_connection.is_connected:
            LOGGER.info("SSH error while opening the connection")
            return False, None, "SSH error Connecting", True

        if sys.stdin.closed:  # pragma: no cover
            # This happened in rare conditions, with previous errors inside Paramiko, which closed stdin.
            LOGGER.debug("stdin was CLOSED, re-opening it")
            sys.stdin = open(0, "r")

        result = self.remote_run('echo $PATH')
        if result is None:  # pragma: no cover  usually, errors are discovered ahead
            # there was a serious error in run, we must abort and remember it
            self.remote_host = None
            self.remote_port = None
            self.ssh_connection = None

            return False, None, "SSH unknown error", True

        if not result.ok:  # pragma: no cover
            LOGGER.info(f"ssh error {result.exited}")
            return False, None, f"SSH error {result.exited}", True
        res = result.stdout.strip()
        if res == "$PATH":  # pragma: no cover : having a Windows server is rare
            return True, "Windows", None, False
        else:
            result = self.remote_run('uname -rsvm')  # noqa

        uname = result.stdout.strip()
        sftp = self.ssh_connection.sftp()
        af = sftp.listdir('.')
        if self.tmp_dir in af:
            self.remote_rmtree(self.tmp_dir)
        sftp.mkdir(self.tmp_dir)
        self.remote_uname = uname
        return True, uname, None, False

    # ---------------------------------------------------------------------------------------------------
    # _________________________________ sql and md5
    #

    def connect_sql(self, config: dict):
        """
        :param config: configuration
        :return: sql_connection, name of 1st album, error
        """
        if (config['sql-host'] is None or config['sql-port'] is None or config['sql-user'] is None
                or config['sql-pwd'] is None) or not config['sql-database'] or not config['enable-database'] \
                or config['dryrun']:
            return None, None, "No SQL configuration"

        if self.sql_connection and self.sql_first_album:
            return self.sql_connection, self.sql_first_album, None

        self.sql_host = config['sql-host']
        self.sql_port = int(config['sql-port'])
        self.sql_user = config['sql-user']
        self.sql_pwd = config['sql-pwd']
        self.sql_database = config['sql-database']
        self.sql_first_album = config['piwigo-album-name']
        LOGGER.debug(f"connect SQL '{self.sql_user}@{self.sql_host}:{self.sql_port}'")
        if self.sql_connection is None:
            try:
                conn = mariadb.connect(
                    user=self.sql_user,
                    password=self.sql_pwd,
                    host=self.sql_host,
                    port=self.sql_port,
                    database=self.sql_database)
            except mariadb.Error as e:
                LOGGER.info(f"Error connecting to MariaDB Platform: {e}")
                self.sql_connection = None
                return None, None, e
            except NameError as e:
                LOGGER.info(f"Error connecting to MariaDB Platform: {e}")
                self.sql_connection = None
                return None, None, e
            self.sql_connection = conn

        cur = self.sql_connection.cursor()

        if self.sql_first_album is None:
            # we need to look for it
            try:
                cur.execute("""SELECT name, id, dir, id_uppercat, uppercats, global_rank, rank
                FROM piwigo_categories 
                WHERE global_rank=1""")  # noqa
            except mariadb.Error as e:
                raise PwpConfigError(f"Error {e} : get name of 1st album, something is wrong in your database ")

            for name, sql_id, path, upper, upper_cats, global_rank, rank in cur:
                self.sql_first_album = name
                self.sql_first_album_id = sql_id
                self.sql_first_album_path = path
                self.sql_first_album_upper_cat = upper
                self.sql_first_album_global_rank = global_rank
                self.sql_first_album_upper_cats = upper_cats
                self.sql_first_album_rank = rank

                return self.sql_connection, name, None
            raise PwpConfigError("No Sql connection and album with global_rank=1, please set album-name")

        # here, album-name is set
        try:
            cur.execute("SELECT name, id, dir, id_uppercat, uppercats, global_rank, rank \n"+  # noqa
                        "FROM piwigo_categories \n" +
                        f"WHERE name='{self.sql_first_album}'")  # noqa
        except mariadb.Error as e:
            raise PwpConfigError(f"Error {e} : get id of 1st album '{self.sql_first_album}' ")

        for name, sql_id, path, upper, upper_cats, global_rank, rank in cur:
            if ',' not in upper:
                # this is really a first level album
                self.sql_first_album_id = sql_id
                self.sql_first_album_path = path
                self.sql_first_album_upper_cat = upper
                self.sql_first_album_global_rank = global_rank
                self.sql_first_album_upper_cats = upper_cats
                self.sql_first_album_rank = rank
                return self.sql_connection, name, None

        raise PwpConfigError(f"album-name argument '{self.sql_first_album}' is not known in piwigo database")

    @staticmethod
    def compute_md5(filename):
        h = hashlib.new('md5')
        with open(filename, "rb") as f:
            size = 2048
            while size == 2048:
                data = f.read(2048)
                size = len(data)
                h.update(data)
            md5 = h.hexdigest()
        return md5

    def build_sql_name(self, filename):
        if not self.sql_first_album:
            raise PwpError("Trying to access database without proper first album set")
        if filename[0] != '/':
            raise PwpError(f"db file name not starting with '/' : '{filename}' ")

        return "./galleries/" + self.sql_first_album + filename

    def sql_get_file_info(self, filename: str) -> FileInfo or None:
        """
        get file information from the sql database, including md5
        :param filename: the path of the file, relative to album
        :return: FileInfo or None
        if not found, id is None
        """
        if not self.sql_connection:
            return None

        LOGGER.debug(f"SQL get information '{filename}'")

        cur = self.sql_connection.cursor()
        sql_name = self.build_sql_name(filename)
        try:
            cur.execute("SELECT id, file, date_available, date_creation, name, author, filesize, " +  # noqa
                        " width, height, date_metadata_update, path, storage_category_id, level, " +
                        " md5sum, added_by, latitude, longitude, representative_ext, lastmodified \n" + # noqa
                        " FROM piwigo_images\n" +
                        f" WHERE path='{sql_name}'")
        except mariadb.Error as e:
            raise PwpError(f"Error {e} : get file info of {sql_name} ")

        if cur.affected_rows > 1:
            raise PwpError(f"SQL Error: get db info from {sql_name} produced {cur.affected_rows} affected row")

        for file_id, file, date_available, date_creation, name, author, file_size, width, height, \
                date_metadata_update, path, storage_category_id, level, md5sum, added_by, latitude, longitude, \
                representative_ext, last_modified in cur:
            if path == sql_name:
                return FileInfo(file_id=file_id, file=file,  date_available=date_available, date_creation=date_creation,
                                name=name, author=author, file_size=file_size, width=width, height=height,
                                date_metadata_update=date_metadata_update, path=path,
                                storage_category_id=storage_category_id, level=level,
                                md5sum=md5sum, added_by=added_by, latitude=latitude, longitude=longitude,
                                representative_ext=representative_ext,
                                last_modified=last_modified)
            raise PwpError(f"""SQL Error: db already has file '{path}' with char case different from '{sql_name}'
                Please change the picture name to match what is already in database""")
        return None

    def sql_set_data(self, filename, md5, size, width, height, latitude, longitude, author,
                     warn_if_no_change=True):
        if not self.sql_connection:
            return False

        width = width or 'NULL'
        height = height or 'NULL'
        LOGGER.debug(f"SQL SET information '{filename}'")

        cur = self.sql_connection.cursor()
        sql_name = self.build_sql_name(filename)

        longitude = longitude or "NULL"
        latitude = latitude or "NULL"
        author = author or "NULL"
        date = datetime.datetime.now()  # self.format_for_sql()
        date_update = f"{date.year}-{date.month:02}-{date.day:02}"

        representative_ext = "NULL" if pathlib.Path(filename).suffix == '.jpg' else 'jpg'

        try:
            cur.execute("UPDATE piwigo_images\n" +
            f" SET md5sum ='{md5}', width={width}, height={height}, filesize={size},\n" +  # noqa
            f" latitude={latitude}, longitude={longitude}, author='{author}', date_metadata_update='{date_update}',\n"
            f" representative_ext='{representative_ext}' "
            f" WHERE path='{sql_name}' """)
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"Error {e} : setting db info to {sql_name} ")

        if cur.affected_rows == 0 and warn_if_no_change:  # pragma: no cover
            # defensive code that should not happen in normal conditions
            # but, nevertheless, no real need to stop processing
            LOGGER.warning(f"Setting db info to {sql_name}: no effect, values where already set to same value")
            return False
        if cur.affected_rows > 1:  # pragma: no cover
            # defensive code that should not happen in normal conditions
            # but, nevertheless, no real need to stop processing
            raise PwpError(f"SQL Error: setting db info to {sql_name} produced {cur.affected_rows} affected row")
        return True

    def sql_get_user_id(self, username: str):
        if not self.sql_connection:
            # not self.sql_connection is supposed to have been trapped ahead
            raise PwpInternalError("sql_get_user_id without SQL connection")

        LOGGER.debug(f"SQL get user id '{username}'")
        cur = self.sql_connection.cursor()

        try:
            cur.execute(f"""SELECT id, username 
            FROM piwigo_users 
            WHERE username='{username}'""")
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"Error {e} : get user id '{username}'")
        if cur.affected_rows > 1:
            raise PwpError(f"SQL Error: get user id '{username}' produced {cur.affected_rows} affected row")

        for index, name in cur:
            if name == username:
                return index
        raise PwpConfigError(f"sql_get_user_id({username}) not found. Change 'piwigo-user' in config")

    def sql_insert_file(self, real_file: str, sql_filename: str) -> FileInfo or None:
        """
        insert sql_filename in the sql database
        :param real_file: the path of the file to be inserted
        :param sql_filename: the path of the file, relative to album
        :return: sql_get_file_info()
        """
        if not self.sql_connection:
            return None  # pragma: no cover    : defensive code
        LOGGER.debug(f"SQL insert file '{sql_filename}'")

        file = os.path.basename(sql_filename)

        date_available = datetime.datetime.now()  # self.format_for_sql()
        date_creation = datetime.datetime.fromtimestamp(os.path.getmtime(real_file))
        name = file
        file_size = int(os.path.getsize(real_file) / 1024)
        path = self.build_sql_name(sql_filename)

        father = os.path.dirname(sql_filename)
        father_info = self.sql_get_dir_info(father, do_insert=True)

        if father_info is None:
            raise PwpError(f"SQL Error: insert file '{sql_filename}' father '{father}' does not exist in db")

        level = self.piwigo_level
        added_by = self.sql_get_user_id(self.piwigo_user)

        # width and height will be inserted afterward, with author and other metadata
        representative_ext = '' if pathlib.Path(file).suffix == '.jpg' else 'jpg'

        src_md5 = ACTOR.compute_md5(real_file)

        cur = self.sql_connection.cursor()

        # NB: other fields in the DB depend on the analysis of the file: jpg, mp4, etc...
        # inserting these values is done elsewhere
        date = datetime.datetime.now()  # self.format_for_sql()
        date_update = f"{date.year}-{date.month:02}-{date.day:02}"
        try:
            cur.execute("INSERT INTO piwigo_images\n" +
                f'SET file="{file}", date_available="{date_available}", date_creation="{date_creation}",\n' +
                f' name="{name}", filesize={file_size}, path="{path}", \n' + # noqa
                f" storage_category_id={father_info.dir_id}, level={level}, date_metadata_update='{date_update}',\n" + # noqa
                f" added_by={added_by}, md5sum='{src_md5}', representative_ext='{representative_ext}'")
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : insert file '{sql_filename}'")
        if cur.affected_rows > 1:
            raise PwpError(f"SQL Error: insert file '{sql_filename}' produced {cur.affected_rows} affected row")

        all_files = self.sql_get_dir_file_list(father_info)
        if file not in all_files:
            raise PwpError(f"SQL Error: insert file '{sql_filename}' not inserted")

        sql_file_info: FileInfo = all_files[file]
        try:
            cur.execute(f"""INSERT INTO piwigo_image_category
                   SET image_id={sql_file_info.file_id}, category_id={father_info.dir_id} 
                   """)
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : insert file category_id'{sql_filename}'")

        if cur.affected_rows > 1:
            raise PwpError(f"SQL Error: set file category_id '{sql_filename}' produced {cur.affected_rows} row")

        return sql_file_info

    #   images: piwigo_image_category:
    #       image_id:                                                                                   yes
    #       category_id:                                                                                yes
    #       rank:               SET ONLY for the picture that are representative of album <> from their dir.
    #       SELECT * FROM `piwigo_image_category` WHERE rank <> 'null'
    #           image_id = 15239, rank = 1, category_id = 1
    #           this is 2006-04-29-07h57-05-Aghios Nikolaios et Route.jpg                                                 # noqa
    #           representative of album

    #   directories: piwigo_categories
    #       id:                                                                             3
    #       name: defaults to dirname, without path                                      2003-07-Juillet-28-bord de mer   # noqa
    #       id_uppercat: null for root(i.e. photos), 1 for first level dir etc...          2 (id of 2003)                 # noqa
    #       comment:                                                                        NULL
    #       dir: dirname, without path                                                   2003-07-Juillet-28-bord de mer   # noqa
    #       rank: index in the enclosing dir, starting at 1                                 31
    #       status: private or public, defaults to private                                  private
    #       site_id: index of the root album, defaults to 1                                 1
    #       visible: true/false, defaults to true                                           true
    #       representative_picture_id: defaults to null                                     271
    #       uppercats: list of categories in path, separated by ',',                        1,2,3                         # noqa
    #       commentable: true/false, defaults to true                                       true
    #       global_rank: list of ranks, separated by '.', starting from root.               1.12.31
    #                    builds a strict order of directories
    #       image_order: defaults to NULL                                                   NULL
    #       permalink:  defaults to NULL                                                    NULL
    #       lastmodified : automatically set                                                2024-01-08 22:04:45           # noqa

    def sql_get_dir_info(self, sql_path, father_info=None, do_insert=False) -> DirInfo or None:
        """
        sql_get_dir_info(self, sql_path, father_path="", father_id=1)

        :param sql_path: the path of the directory inside the father. if father is None, from root after photo
        :param father_info: info of the father.
        :param do_insert: if True, inserts the dir when absent
        :return: DirDescr
        """
        if not self.sql_connection:
            return None                  # pragma: no cover: defensive code

        LOGGER.debug(f"SQL get dir id '{father_info}'/'{sql_path}'")

        if sql_path[0] == '/':
            sql_path = sql_path[1:]

        if '/' in sql_path:
            dir_name = sql_path.split('/')[0]
            path_len = len(dir_name) + 1
            next_path = sql_path[path_len:]
        else:
            dir_name = sql_path
            next_path = ""

        if father_info is None:
            father_info = DirInfo(name=self.sql_first_album,
                                  path=self.sql_first_album,
                                  dir_id=self.sql_first_album_id,
                                  rank=self.sql_first_album_rank,
                                  id_upper_cat=self.sql_first_album_upper_cat,
                                  upper_cats=self.sql_first_album_upper_cats,
                                  global_rank=self.sql_first_album_global_rank)

        sql_sons_descr = ACTOR.sql_get_dir_sub_dirs(father_info)

        if dir_name in sql_sons_descr.keys():
            son = sql_sons_descr[dir_name]
        elif do_insert:
            son = self.sql_insert_dir_and_reorder(dir_name, father_info, sql_sons_descr)
        else:
            # dir_name was not found in father
            return None

        assert son, f"directory {dir_name} inserted incorrectly in {father_info}"
        if next_path == "":
            # we have reached the last item
            return son
        return self.sql_get_dir_info(next_path, son, do_insert=do_insert)

    def sql_insert_dir_at_rank(self, dir_name: str, father_info: DirInfo, rank) -> DirInfo or None:
        """
        inserts dir_name inside father, which is not None
        Also reorders sons of father in increasing order
        CAVEAT: Does NOT check if other sons (in db) are also valid subdirectories (i.e. dir exist in filesystem)
        CAVEAT: does NOT reorder the sons

        :param dir_name: dir-name, without the full path
        :param father_info:
        :param rank: new rank
        :return: DirInfo
        """

        if not self.sql_connection:
            return None
        LOGGER.debug(f"SQL insert dir {dir_name} in '{father_info}' at rank {rank} ")

        if not father_info:
            raise PwpInternalError(f"NULL father for {dir_name}, rank={rank}")
        cur = self.sql_connection.cursor()
        try:
            # CAVEAT: upper_cats ends with the directory OWN ID, that we do not know until it is created !
            global_rank = str(father_info.global_rank) + '.' + str(rank)
            cur.execute("INSERT INTO piwigo_categories \n" +
                        f' SET name="{dir_name}", id_uppercat="{father_info.dir_id}", \n' +  # noqa
                        f' dir="{dir_name}", rank={rank}, status="private", site_id=1, \n' +
                        ' visible="true", commentable="true", \n' +
                        f' global_rank="{global_rank}" ')  # image_order="NULL", permalink="null"
            # NB: RETURNING does not seem to be supported on all mariaDB implementations.
            # NB: uppercats=CONCAT("{father_info.upper_cats}", ",", id), does not work, id=0 # noqa
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e}: INSERT dir {dir_name} in {father_info}")
        if cur.affected_rows != 1:
            raise PwpError(f"SQL Error: {cur.affected_rows} affected rows inserting dir {dir_name} ")

        info = self.sql_get_dir_info(dir_name, father_info=father_info, do_insert=False)
        if info is None:
            raise PwpInternalError(f"SQL error dir {dir_name} not correctly inserted")

        upper_cats = str(father_info.upper_cats) + ',' + str(info.dir_id)
        try:
            cur.execute(" UPDATE piwigo_categories \n" +
                        f' SET uppercats="{upper_cats}" \n' +  # noqa
                        f" WHERE id={info.dir_id}")
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e}: UPDATING uppercats for dir {dir_name} ")  # noqa

        info = self.sql_get_dir_info(dir_name, father_info=father_info, do_insert=False)
        # we verify upper cats were inserted correctly
        if info is None:
            raise PwpInternalError(f"SQL error dir {dir_name} not correctly inserted after update of upper cats")

        LOGGER.msg(f"inserted dir '{dir_name}:{info.dir_id}' into database, uppercats = '{upper_cats}'")   # noqa
        return info

    def sql_insert_dir_and_reorder(self, dir_name, father_info: DirInfo, sql_sons_descr) -> DirInfo:
        """
        inserts dir_name inside father, which is not None
        Also reorders sons of father in increasing order
        CAVEAT: Does NOT check if other sons (in db) are also valid subdirectories (i.e. dir exist in filesystem)

        :param dir_name: dir-name, without the full path
        :param father_info:
        :param sql_sons_descr:
        :return: DirInfo
        """

        file_list = list(sql_sons_descr.keys())
        file_list.append(dir_name)
        file_list.sort()
        info = None
        rank = 1
        for file in file_list:
            if file == dir_name:
                info = self.sql_insert_dir_at_rank(dir_name, father_info, rank)
            else:
                son_info = sql_sons_descr[file]
                self.sql_change_dir_rank(father_info, son_info, rank)
            rank += 1

        return info

    def sql_change_dir_rank(self, father_info: DirInfo, son_info: DirInfo, new_rank):
        if not self.sql_connection:
            return None                  # pragma: no cover: defensive code
        old_rank = son_info.rank
        if new_rank == old_rank:
            return son_info

        LOGGER.debug(f"SQL change dir {son_info} rank in '{father_info}'  {son_info.rank} -> {new_rank} ")

        le = len(str(old_rank)) + 1
        new_global_rank = str(son_info.global_rank)[:-le] + '.' + str(new_rank)
        cur = self.sql_connection.cursor()

        try:
            cur.execute("UPDATE piwigo_categories\n" +
                   f"SET rank='{new_rank}', global_rank='{new_global_rank}' \n" +
                   f"WHERE dir='{son_info.name}' and id_uppercat={father_info.dir_id}")           # noqa
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : change dir {son_info} rank {old_rank} -> {new_rank} in {father_info}")
        if cur.affected_rows != 1:
            raise PwpError(f"SQL Error: {cur.affected_rows} affected rows, change dir {son_info} rank in {father_info}")
        return True

    def sql_get_dir_file_list(self, dir_info: DirInfo) -> dict[str, FileInfo]:
        """
        sql_get_dir_file_list(self, path):
        :param dir_info: dir we are looking in the database
        :return: dictionary of files: result[basename]= FileInfo
        """
        if self.sql_connection is None or self.sql_connection is False:
            return {}                 # pragma: no cover: defensive code

        if dir_info is None:
            return {}

        LOGGER.debug(f"SQL get dir files '{dir_info}'")

        cur = self.sql_connection.cursor()

        try:
            cur.execute("SELECT id, file, date_available, date_creation, name, author, filesize, " +  # noqa
                        " width, height, date_metadata_update, path, storage_category_id, level, " +
                        " md5sum, added_by, latitude, longitude, representative_ext, lastmodified \n" +                   # noqa
                        " FROM piwigo_images\n" +
                        f" WHERE storage_category_id={dir_info.dir_id}")
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"Error {e} : getting sql files of dir {dir_info}")

        result = {}
        for file_id, file, date_available, date_creation, name, author, file_size, width, height, \
                date_metadata_update, path, storage_category_id, level, md5sum, added_by, latitude, longitude, \
                representative_ext, last_modified in cur:
            father = os.path.dirname(path)
            if father != './galleries/' + dir_info.path:
                raise PwpError(f"incoherent father '{father}' for file {file}:{id}")
            result[file] = FileInfo(file_id=file_id, file=file, date_available=date_available,
                                    date_creation=date_creation,
                                    name=name, author=author, file_size=file_size, width=width, height=height,
                                    date_metadata_update=date_metadata_update, path=path,
                                    storage_category_id=storage_category_id, level=level,
                                    md5sum=md5sum, added_by=added_by, latitude=latitude, longitude=longitude,
                                    representative_ext=representative_ext,
                                    last_modified=last_modified)

        return result

    def sql_get_dir_sub_dirs(self, dir_info: DirInfo):
        """
        sql_get_dir_sub_dirs(self, path):
        :param dir_info: info about dir we are looking in the database
        :return: dictionary of subdirectories: result[basename]= DirInfo
        """
        if not self.sql_connection:
            return {}                 # pragma: no cover: defensive code

        if dir_info is None:
            return {}

        LOGGER.debug(f"SQL get dir sub-dirs '{dir_info}'")

        cur = self.sql_connection.cursor()

        try:
            cur.execute("SELECT id, dir, rank, id_uppercat, global_rank, uppercats \n" +  # noqa
            " FROM piwigo_categories \n" +
            f" WHERE id_uppercat={dir_info.dir_id}")      # noqa
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : getting sql files of dir {dir_info}")

        result = {}
        for sql_id, sql_dir, sql_rank, sql_upper, sql_global, upper_cats in cur:
            new_path = dir_info.path + '/' + sql_dir
            result[sql_dir] = DirInfo(name=sql_dir, path=new_path, dir_id=sql_id, rank=sql_rank,
                                      id_upper_cat=sql_upper, global_rank=sql_global, upper_cats=upper_cats)
        return result

    def sql_get_dir_category_content(self, dir_info: DirInfo):
        if not self.sql_connection:
            return None                 # pragma: no cover: defensive code

        cur = self.sql_connection.cursor()
        try:
            cur.execute(f"""SELECT image_id, rank  
                        FROM piwigo_image_category             
                        WHERE category_id={dir_info.dir_id}""")
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : getting dir {dir_info} content as a category")

        result = {}
        for sql_id,  rank in cur:
            result[sql_id] = {'id': sql_id, 'rank': rank}
        return result

    def sql_remove_file_from_db(self, p_config: dict, file_info: FileInfo):
        if not self.sql_connection:
            return True  # pragma: no cover: defensive code

        LOGGER.msg(f"Database: Remove {file_info}")
        real_path = p_config["album"] + '/' + file_info.path

        if self.isfile(real_path):
            raise PwpError(f"Error: file {real_path} prevents from removing {file_info} from database")

        cur = self.sql_connection.cursor()
        try:
            cur.execute(f"""DELETE FROM piwigo_images             
            WHERE id={file_info.file_id}""")

            cur.execute(f"""DELETE FROM piwigo_image_category             
            WHERE image_id={file_info.file_id}""")

            self.sql_connection.commit()

        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : removing file {file_info} ")

        if cur.affected_rows != 1:
            raise PwpError(f"SQL Error: {cur.affected_rows} affected rows removing {file_info}'")
        return True

    def sql_remove_dir_from_db(self, dir_info: DirInfo):
        """
        sql_remove_dir_from_db(self, dir_info: DirInfo):
        :param dir_info:  directory to remove from DB
        :return: True

        CAVEAT: if the directory is NOT empty, use sql_remove_dir_from_database_recursive
        """

        if not self.sql_connection:
            return False

        # first, test that all pictures and subdirectories have been deleted from database

        LOGGER.msg(f"Database: Remove {dir_info}")

        sql_file_descr = ACTOR.sql_get_dir_file_list(dir_info)
        if sql_file_descr:
            for file in sql_file_descr:
                LOGGER.msg(f"found file {file}")
            raise PwpError(f"SQL Error removing directory {dir_info} : still holds files in DB")

        sql_file_descr = ACTOR.sql_get_dir_category_content(dir_info)
        if sql_file_descr:
            for file in sql_file_descr:
                LOGGER.msg(f"found file {file}")
            raise PwpError(f"SQL Error removing category {dir_info} : still holds files in DB")

        sql_sons_descr = ACTOR.sql_get_dir_sub_dirs(dir_info)
        if sql_sons_descr:
            for file in sql_sons_descr:
                LOGGER.msg(f"found subdir {file}")
            raise PwpError(f"SQL Error removing directory {dir_info} : still holds directories in DB")

        # then, do remove

        cur = self.sql_connection.cursor()
        try:
            cur.execute(f"""DELETE FROM piwigo_categories             
                 WHERE id={dir_info.dir_id}""")
            self.sql_connection.commit()
        except mariadb.Error as e:
            raise PwpError(f"SQL Error {e} : removing directory {dir_info}")

        if cur.affected_rows != 1:
            nb = cur.affected_rows
            raise PwpError(f"SQL Error: {nb} affected rows removing directory {dir_info}")
        return True

    def sql_remove_dir_from_database_recursive(self, p_config: dict, dir_info: DirInfo):
        """
        sql_remove_dir_from_database_recursive
        :param p_config:
        :param dir_info: dir to remove
        :return: True

        Called when a directory is empty on the file system, but still has items in the database
        recursively deletes all sons,
        then deletes the dir itself

        while deleting the sons (directory and file), it is verified that these do not exist anymore in the filesystem
        so this method is safe
        """
        sql_file_descr = ACTOR.sql_get_dir_file_list(dir_info)
        for file in sql_file_descr:
            ACTOR.sql_remove_file_from_db(p_config, sql_file_descr[file])
            # will yell if the file still exists

        sql_sons_descr = ACTOR.sql_get_dir_sub_dirs(dir_info)
        for subdir in sql_sons_descr:
            self.sql_remove_dir_from_database_recursive(p_config, sql_sons_descr[subdir])
            # will yell if there are still files, but NOT if there are empty directories

        self.sql_remove_dir_from_db(dir_info)

    # def trace_malloc_start(self):
    #     gc.enable()
    #     if not self.trace_malloc:
    #         return
    #     tracemalloc.start()      # noqa
    #
    # def trace_malloc_snapshot(self, name: str, snapshot1: tracemalloc.Snapshot or None = None, garbage=False):     # noqa
    #     if garbage:
    #         gc.collect(2)
    #     if not self.trace_malloc:
    #         return None
    #     LOGGER.msg("*** start gc garbage ***")
    #     LOGGER.msg(gc.garbage)
    #     LOGGER.msg("*** end gc garbage ***")
    #     snapshot2 = tracemalloc.take_snapshot()       # noqa
    #     if snapshot1:
    #         top_stats = snapshot2.compare_to(snapshot1, 'lineno')
    #         LOGGER.msg(f"{name} : [ Top 10 differences ]")
    #         for item in top_stats[:10]:
    #             LOGGER.msg(item)
    #     return snapshot2


ACTOR = PwpActor()



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------
import platform
import sys
import os
import pprint

from piwiPre.pwpErrors import LOGGER
from piwiPre.pwpArgsIni import PwpArgsIni


class PwpParser(PwpArgsIni):
    default_ini_file = "piwiPre.ini"

    def __init__(self, arguments=None, with_config=True, program: str = "piwiPre"):
        super().__init__()
        self.config = None  # the config after parsing HOME,& cwd ini files and cmdline args

        self.add_header("""
.. _configuration:
        
Commandline Flags and configuration items
#########################################

This file is the default configuration of piwiPre.

Unless stated otherwise, the  configuration items have a command line argument counterpart,
with the same name, starting with - - .

The default value is given as an argument.

The configuration file uses the yaml syntax,
and uses pyYaml  to read/write the configuration file

- *boolean* values are *true* and *false*
- *None* denotes a value which is not set.
- *string* SHOULD single or double quotes to prevent yaml to interpret values.
- *directory* should be a valid path syntax (absolute or relative), written as a string.
- *dictionary* read key : value
""")
        self.add_header("""
Drapeaux de ligne de commande et éléments de configuration
##########################################################

Ce fichier est la configuration par défaut de piwiPre.

Sauf lorsque indiqué autrement, les éléments de configuration ont une contrepartie sur la ligne de commande,
avec le même nom, mais commençant par - - .

La valeur par défaut est indiquée plus bas.

Le fichier de configuration utilise la syntaxe de YAML,
et piwiPre utilise pyYaml pour lire/écrire la configuration.

- les valeurs *booléennes*   sont *true* et *false* (vrai et faux)
- *None* dénote une valeur qui est vide
- *string* DOIT être une chaîne de caractères délimitée par les caractères ' ou \"
- *directory* doit être ue syntaxe de chemin valide (absolut ou relatif), écrit comme une chaîne.
- un dictionnaire (*dictionary*) s'écrit: clef : valeur         
        """, lang='fr')

        self.add_header("""

Configuration hierarchy
=======================

1. Default values are set for all items.

2. By default, configuration data is read from files named 'piwiPre.ini',
   but if the cmdline arguments hold a '--ini-file new-value.ini', then the name of the ini file is changed,
   so that the new ini-file is taken into account.

3. In the user HOME directory, as a special case, '.piwiPre.ini' is read instead.
   This file should be protected against reading from others,
   (chmod 500 in the Linux case).
   It is used to store confidential information:

   - SSH information : remote-user, remote-host, remote-port
   - NB: SSH password is stored differently see :ref:`ssh`
   - SQL information: sql-host, sql-port, sql-user, sql-password
   - piwigo information: piwigo-user
   
   This file also typically stores **default** directories:

   - triage, album, web
   - remote-web and enable-remote-web

   Other global configuration items could also be stored there:
   
   - month-name, copyright, instructions, names, authors, dates
  
4. in cwd. 

   This cwd/piwiPre.ini should store **non-default** directories and information, 
   that are specific to this particular directory
   
   So, if the user wants to maintain various settings, this can be done in a per directory basis,
   
   For instance one directory with 1 naming convention and another directory for a different naming
   Here, one usually  sets up the global configuration without the confidential information
   and without details specific to each TRIAGE directory.
   
   If this file stores confidential information, it should also be chmod 500.
   
   CAVEAT: Once cwd/piwiPre.ini has been read, new values of the confidential configuration 
   are no more taken into account.

5. On cmdLine. WHen used on the command-line, options start with '--', such as '--version' etc.
 
6. When managing TRIAGE, in TRIAGE subdirectories. 

   These .ini files are read only when processing files in TRIAGE.
   Only directory-specific configuration should be stored here.
   
   To clarify a difficulty: when managing TRIAGE, the configuration files in WEB are *not* read
     
   Typically, one stores there 'names', 'authors', 'dates', 'copyright', 'instructions',
   if some of these should be different for a specific directory.
   
   If enable-auto-configuration is true, this file will be copied in the corresponding WEB directory
   
   If there was a preexisting .ini file in the WEB subdir, then it is clobbered by the new one.
   

7. When managing ALBUM, in the directory hierarchy of WEB.

   These .ini files are read only when processing files in ALBUM.
   
   .ini files in WEB are usually a copy of an .ini file in the original TRIAGE directory, but they *can* be
   hand-writen by the piwigo administrator.
    
   To clarify a difficulty: when managing ALBUM, the configuration files in TRIAGE are *not* read
    
   For instance, some sub-ALBUM may hold pictures without shooting date in the Exif data,
   therefore the naming method is different.
   
NB 1: The later has precedence over the sooner.
 
NB 2: Therefore, cmdLine items do not modify configuration options found in directories of TRIAGE and WEB.
   
   The only way to reset these are:
   
   - To modify the .ini files in TRIAGE, and then run piwiPre to forward the modifications to WEB
   - To edit .ini files in WEB 
   
.. attention::

   piwiPre.ini and HOME/.piwiPre.ini should be written with UTF8 encoding

""")

        self.add_header("""

Hiérarchie des configurations
=============================

La configuration de piwiPre est calculée dans l'ordre suivant:

1) Chaque item de configuration se voit attribuer sa valeur par défaut, telle qu'indiquée ici.

2) Par défaut, la configuration sera lue dans des fichiers nommés 'piwiPre.ini', 
   mais si la ligne de commande contient un élément de type '--ini-file nouveau-nom.ini',
   alors ce sont les fichiers nouveau-nom.ini qui seront utilisés.
   
3) Dans le répertoire personnel de l'utilisateur (HOME), c'est le fichier .piwiPre.ini qui est lu. 
   noter le '.' initial, qui dénote un fichier 'caché'
   
   Ce fichier DOIT être protégé contre la lecture par d'autres personnes que son propriétaire 
   (l'équivalent de chmod 500 pour le cas linux), car il contient des informations confidentielles:
   
   - sql-host sql-port sql-user sql-pwd
   - remote-host remote-port  remote-user remote-pwd remote-incoming
   - piwigo-user 
  
   D'autres éléments de configuration peuvent aussi être stockés dans HOME/.piwiPre.ini.


4. Dans cwd, c'est à dire le répertoire d'où est lancé piwiPre.

   Ici, on mets habituellement en place la configuration globale pour un album racine particulier, sans les 
   informations confidentielles qui sont dans HOME, et sans les détails spécifique à chaque sous-répertoire de TRIAGE.
   
   Donc ce fichier est en général vide quand n seul album racine ('photo', par défaut) est utilisé.
   
   Si ce répertoire contient des informations confidentielles 
   (par exemple dans un cas où un album racine n'est pas accédé de la même façon que les autres)
   il doit alors, lui aussi, être protégé en lecture (chmod 500)
   
   Une fois que cqd/piwiPre.ini a été lu, aucune autre information confidentielle supplémentaire ne sera prise en compte
   
5. Sur la ligne de commande. Dans ce cas, les options suivent l'usage Linux, et commencent donc par '--', 
   par exemple '--help'
   
6. lorsque on importe des photos, (et donc TRIAGE est défini, et verify-album non défini), 
   dans chacun des sous-répertoire de TRIAGE.
   
   Ces fichiers .ini ne sont lu que en cas d'importation, c'est-à-dire ne sont pas lus lorsque verify-album est défini.
   
   Typiquement, on va stocker dans ces fichiers des valeurs spécifiques au répertoire pour 
   ‘names’, ‘authors’, ‘dates’, ‘copyright’, ‘instructions’. Le .ini sera alors recopié dans le sous-repertoire de
   WEB pour que ces valeurs soient aussi utilisées lors d'une phase de vérification des albums.
   
   Si un fichier .ini était présent dans le sous-repertoire de WEB il sera écrasé par le nouveau.

7. Lorsque on gère les albums, c'est à dire lorsque verify-album est défini, dans la hiérarchie de WEB.

   Ces .ini peuvent avoir été générés automatiquement à travers enable-auto-configuration, 
   ou bien avoir été écrits manuellement par l'administrateur.
   
   Clarifions un point: quand on gère les albums, les fichiers de configuration dans TRIAGE ne sont *pas* lus.
    
   ces fichiers .ini sont utilisés typiquement pour maintenir des élements de configuration spécifiques aux sous-albums 
   
   Par exemple, un sous-album peut contenir des photos qui n'ont pas de meta-données de date, 
   et donc la methode de gestion de dates est spécifique.


Dans cette hiérarchie de configuration, les dernières étapes ont précédence sur les premières.
Donc les seules façons de changer un .ini dans WEB sont de:

- exécuter une importation de fichier, avec un nouveau .ini dans TRIAGE
- modifier à la main le .ini dans WEB

       
        """, lang='fr')

        self.add_header("""
Some vocabulary
===============

- The **piwigo host** is the server where the piwigo service runs. 
  Usually this is a cloud-based host, or a NAS, or a Linux server.

- The **piwiPre host** is the computer where piwiPre runs. 
  Usually this is a desktop PC, running Linux or Windows, but could be also the same machine than piwigo host.

- **cwd** is the directory where piwiPre is run. If a relative path is used, then it starts from there. 
  For instance, 'album: ALBUM' specifies that the current album directory is cwd/ALBUM
""")

        self.add_header("""
Vocabulaire
===========

- l'**hôte piwigo** est le serveur sur lequel le service piwigo tourne. 
  En général, c'est un serveur dans le cloud, ou un NAS, ou un serveur Linux domestique.
  
- l'**hôte piwiPre** est le calculateur sur lequel piwiPre tourne.
  En général, c'est un PC personnel, sous Windows ou Linux, mais ce pourrait aussi être la même machine
  que l'hôte piwigo.
  
- **cwd** (c'est à dire Current Working Directory, le répertoire de travail courant) est le répertoire dans lequel
  piwiPre s'exécute. Lorsqu'un chemin relatif est spécifié, il est relatif par rapport à cwd.
  Par exemple "album: ALBUM' spécifie que l'album courant est cwd/ALBUM""", lang='fr')

        self.add_header("""
cmdLine flags only
==================

The following command line arguments do not have configuration counterpart in the .ini file:
""")

        self.add_header("""
Les drapeaux de configuration spécifiques de la ligne de commande
=================================================================

Les drapeaux suivants n'ont pas de contrepartie dans les fichiers .ini, 
et ne peuvent donc être utilisés que sur la ligne de commande
""")

        # -h, --help is implicit
        self.add_item('quiet',
                      action='store_true',
                      help="Does not print the initial banner with log information.",
                      fr_help="N'imprime pas la banière initiale",
                      location='args',
                      config="""
        This flag has no value, it is active only when set""",
                      fr_config="""
        Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
        """)

        self.add_item('version',
                      action='store_true',
                      help="Prints piwiPre version number and exits.",
                      fr_help="Imprime la version de piwiPre et sort",
                      location='args',
                      config="""
This flag has no value, it is active only when set
Usually used with --quiet
""",
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
Utilisé habituellement avec --quiet
""")
        self.add_item('full-help',
                      action='store_true',
                      help="Prints piwiPre extended help and exits.",
                      fr_help="Imprime l'aide étendue de piwiPre et sort",
                      location='args',
                      config="""
This flag has no value, it is active only when set""",
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
        """)

        self.add_item('licence',
                      action='store_true',
                      help="prints the LICENCE and exits",
                      fr_help="imprime la LICENCE et sort",
                      location='args',
                      config="""
This flag has no value, it is active only when set""",
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
""")

        self.add_item('debug',
                      action='store_true',
                      help="Increments the level of verbosity of the logs printed on standard output.",
                      fr_help="Incrémente le niveau de verbosité des logs affichés sur la sortie standard.",
                      location='args',
                      config="""
This flag has no value, it is active only when set""",
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
""")

        self.add_item('stop-on-warning',
                      action='store_true',
                      help="Stops piwiPre at the first warning.",
                      fr_help="Stoppe piwiPre au premier avertissement (Warning)",
                      location='args',
                      config="""
        This flag has no value, it is active only when set""",
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
""")

        self.add_item('trace-malloc',
                      action='store_true',
                      help="Uses trace-malloc to look for memory leaks, use at your own risks...",
                      fr_help="Utilise trace-malloc pour rechercher des memory-leaks, reservé aux développeurs",
                      location='args',
                      config="""
This flag has no value, it is active only when set""",
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.
""")

        self.add_item('dump-config',
                      pwp_type=str,
                      action='store',
                      help="Dump the configuration for a given directory and exits.",
                      fr_help="Affiche la configuration pour le repertoire argument, et sort",
                      location='args',
                      config="""
The value of this flag is the name of the directory from which the configuration should be dumped.
  
This path starts from cwd, e.g. TRIAGE/Armor""",
                      fr_config="""
La valeur de ce drapeau est le nom du repertoire dont on veut afficher la configuration.
                      
Le chemin part depuis cwd, par exemple TRIAGE/Armor""")

        self.add_item('dryrun',
                      action='store_true',
                      help="Prints what should be done, but does not execute actions.",
                      fr_help="Affiche les actions que piwiPre devrait faire, mais ne les réalise pas",
                      location='args',
                      config="""this flag has no value, it is active only when set
CAVEAT:

    dryrun tries to display all potential actions that would be made, 
    but there are some limitations. 
    For instance, dryrun does NOT correctly detect all thumbnail-related activities.
                      """,
                      fr_config="""
Ce drapeau n'a aucune valeur, il n'est actif que si utilisé.

ATTENTION:
    dryrun essaie d'afficher toutes les actions potentielles qui pourraient être réalisées,
    mais il y a des limitations à ce qu'il peut découvrir.
    En particulier, dryrun n'affiche pas correctement l'intégralité des actions relatives au miniatures (thumbnails)
""")

        self.add_item('ini-file',
                      help="Changes the default configuration file to something else.",
                      fr_help="Change le nom du fichier de configuration par défaut",
                      pwp_type=str,
                      default='piwiPre.ini',
                      action='store',
                      location='args',
                      config="""
The value of this flag is the name of the new configuration file, it will be used in all directories

**CAVEAT**:

    A common mistake is to set a specific configuration file in an other directory.
    
    For instance, if we want to work with tests\\sources\\piwiPre-Usecase-6.ini inside tests\\results\\Usecase, 
    we could believe we have to run '--ini-file tests\\sources\\piwiPre-Usecase-6.ini', which does not deliver 
    the expected result.
    
    The right way is to:
    
    - copy tests\\sources\\piwiPre-Usecase-6.ini into tests\\results\\Usecase 
    - execute  '--chdir tests\\results\\Usecase --ini-file piwiPre-Usecase-6.ini'
        
    or, even better:
    
    - copy tests\\sources\\piwiPre-Usecase-6.ini as tests\\results\\Usecase\\piwiPre.ini
    - execute  '--chdir tests\\results\\Usecase' 
""",  # noqa
                      fr_config="""
La valeur de ce drapeau est le nom du nouveau fichier de configuration, il sera utilisé dans tous les repertoires.

**ATTENTION**:
 
    Une erreur courante est de croire qu'il s'agit d'indiquer 1 fichier de configuration particulier.
    
    Par exemple, pour travailler avec le .ini de tests\\sources\\piwiPre-Usecase-6.ini dans le repertoire 
    tests\\results\\Usecase, croire qu'il faut faire : '--ini-file tests\\sources\\piwiPre-Usecase-6.ini'
 
    Il faut plutôt :
      
    - recopier tests\\sources\\piwiPre-Usecase-6.ini dans tests\\results\\Usecase 
    - exécuter  '--chdir tests\\results\\Usecase --ini-file piwiPre-Usecase-6.ini'
        
    ou, encore mieux :
    
    - recopier tests\\sources\\piwiPre-Usecase-6.ini dans tests\\results\\Usecase\\piwiPre.ini
    - exécuter  '--chdir tests\\results\\Usecase'
""")  # noqa

        self.add_item('reset-ini',
                      help="Resets to default the ini file in the current directory and exits.",
                      fr_help='Génère in fichier .ini par défaut dans le répertoire courant et sort',
                      config="""
- value = triage    Generates a configuration file *without* piwigo, used for triage only
- value = default   Generates a configuration file with *all* values reset to their default
- value = local     Generates a configuration file suitable for local access to piwigo directories
- value = remote    Generates a configuration file suitable for remote (ssh) access to piwigo directories
- value = None      Does nothing
    
Previous .ini file is renamed piwiPre.ini.bak, piwiPre.ini.bak is renamed piwiPre.ini.bak-1, etc """,
                      fr_config="""
- valeur = triage    Génère un fichier de configuration *sans* piwigo, uniquement pour renommer les fichiers
- valeur = default   Génère un fichier de configuration avec *toutes* les valeurs remises à leur valeur par défaut
- valeur = local     Génère un fichier de configuration pour un accès local aux répertoires de piwigo
- valeur = remote    Génère un fichier de configuration pour un accès distant (ssh) aux répertoires de piwigo
- valeur = None      Ne fait rien

Le fichier .ini précédent est renommé piwi piwiPre.ini.bak, piwiPre.ini.bak est renommé piwiPre.ini.bak-1, etc...
""",
                      action='store',
                      default=None,
                      choices=['triage', 'default', 'local', 'remote', None],
                      location='args')

        self.add_item('chdir',
                      help="Changes the default directory where piwiPre is run, is always executed BEFORE --ini-file",
                      fr_help="Change le répertoire d'où est exécuté piwiPré, toujours effectué AVANT --ini-file",
                      pwp_type=str,
                      action='store',
                      location='args')

        self.add_item('recursive-verify-album',
                      help="Makes --verify-album recursive (go in sub-directories)",
                      fr_help="Rend --verify-album récursif (va dans les sous-répertoires)",
                      action='store',
                      choices=['true', 'false'],
                      default='false')

        self.add_item('restart-from-dir',
                      help="During verify-album, restart from this directory",
                      fr_help='Pendant verify-album, repart de ce sous-repertoire',
                      pwp_type=str,
                      action='store',
                      config="""
Directories to verify are  sorted in alphanumerical order.

Directories less than the argument are not managed.

The argument is the first managed.

If the argument is does not start with the value of --verify-album, then restart-from-dir is ignored,
because there is no chance that this would be a sub-directory.  

If the argument starts with the same value than the value of --verify-album, but the directory is not found,
then an error is raised.


For instance, next line will start verifying at 2012/2012-08-Aout-03-Example

    piwiPre --verify-album 2012 --restart-from-dir 2012/2012-08-Aout-03-Example

but next line will verify 2013
 
    piwiPre --verify-album 2013 --restart-from-dir 2012/2012-08-Aout-03-Example
    
and next line will generate an error if 2012/wrong-subdir does not exist 

    piwiPre --verify-album 2012 --restart-from-dir 2012/wrong-subdir

This flag may be useful to restart processing that was interrupted
    """,
                      fr_config="""
Les répertoires à vérifier sont triés dans l'ordre lexicographique.

Les répertoires 'avant' l'argument ne sont pas vérifiés.

Le répertoire argument est le premier trié.

Si l'argument de --restart-from-dir ne commence pas par la valeur de --verify-album, sa valeur est ignorée, puisque ce ne peut pas
être un sous-répertoire de ALBUM.

Si l'argument de --restart-from-dir n'est pas trouvé dans l'album, alors une erreur est générée.

Par exemple, la ligne suivante démarre la vérification à 2012/2012-08-Aout-03-Example:

    piwiPre --verify-album 2012 --restart-from-dir 2012/2012-08-Aout-03-Example
    
Mais la ligne suivante vérifie 2013 entièrement:

    piwiPre --verify-album 2013 --restart-from-dir 2012/2012-08-Aout-03-Example
    
et la ligne suivante génère une erreur si 2012/wrong-subdir n'existe pas  

    piwiPre --verify-album 2012 --restart-from-dir 2012/wrong-subdir
                         
""")  # noqa

        self.add_header("""
Global actions in ALBUM subdirectories
======================================
""")
        self.add_header("""
Actions globales sur les sous-répertoires de ALBUM
==================================================
""", lang='fr')

        self.add_item('verify-album',
                      help='Directory in ALBUM to be verified, use it multiple times to add more albums',
                      fr_help="Repertoire de ALBUM à vérifier, à utiliser plusieurs fois pour ajouter d'autres valeurs",
                      action='append',
                      default=[],
                      pwp_type=list,
                      config="""
   
- Value = a directory in ALBUM to be verified
- Default : [].
- may be used several times
- '*' is a special value, it means : all subdirectories of the root album, (provided --recursive-verify-album is set) 

If verify-album is set, triage is unset

Caveat: sub-directories of the target directory are NOT verified, 
unless --recursive-verify-album is set, which is not the default

Other useful flags with their typical value when verifying albums:

        --restart-from-dir folder/sub-sub-dir      # (just in case this is necessary)
        --recursive-verify-album true              # useful only if folder1 or folder2 have sub-folders 
        --enable-thumbnails true                   # build thumbnails if they were not built
        --enable-thumbnails-delete true            # remove useless thumbnails
        --enable-metadata true                     # if metadata is lacking, will be set
        --enable-rotation true                     # rotate the pictures 
        --enable-database true                     # set in database
        --enable-conversion true                   # change pictures to jpg and video to mp4    

        --enable-metadata-reset false              # trust metadata that was generated previously
        --trust-date-in-filename true              # trust date that was generated previously 
        --enable-rename false                      # trust names that were generated previously
""")

        self.add_item('enable-thumbnails-delete',
                      help='Enables deletion of useless piwigo thumbnails.',
                      fr_help="Autorise l'enlèvement des miniatures (thumbnails) piwigo inutiles",
                      action='store',
                      choices=['true', 'false'],
                      default='false',
                      config="""
When doing verify-album  
this flag allows to remove thumbnails that are useless because there is no corresponding picture.
   
- It should be tested first with --dryrun
- For security, the default value is 'false': the user AS TO set explicitly to 'true'
        """,
                      fr_config="""
Dans une passe de vérification des albums (verify-album),
ce drapeau autorise l'enlèvement des miniatures piwigo qui sont devenues obsolete car correspondant
à une image qui n'existe plus.

- Il est prudent de tester son effet avec --dryrun
- Par sécurité, sa valeur par défaut est 'false', l'utilisateur DOIT le positionner explicitement à 'true' 
""")

        self.add_header("""
.. attention::
    The following flags --test-xxx are used when performing self-testing of piwiPre
    They are not intended to be used under normal circumstances""")

        self.add_header("""
.. attention::
    Les drapeaux suivants, de la forme --test-XXX sont utilisés lors de l'autotest de piwiPre,
    et ne sont donc pas sensés être utilisés dans des circonstances habituelles        
""", lang='fr')

        self.add_item('test-ssh',
                      help="tests ssh on remote host and exits",
                      fr_help="teste la communication ssh avec l'hôte distant et sort",
                      action='store_true',
                      location='args')

        self.add_item('test-sftp',
                      help="tests sftp on remote host (by copying a file in HOME) and exit",
                      fr_help="teste sftp avec l'hôte distant (en copiant un fichier dans HOME) et sort",
                      action='store_true',
                      location='args')

        self.add_item('test-sql',
                      help="tests SQL access on sql host by looking at first picture",
                      fr_help="teste l'accès SQL avec le serveur en atteignant la 1ere image",
                      action='store_true',
                      location='args')

        self.add_header("""
Management of directories
=========================""")

        self.add_header("""
Gestion des répértoires
=======================""", lang='fr')

        self.add_item('triage',
                      help='Sets the TRIAGE directory where are stored incoming pictures/videos.',
                      fr_help="Indique le nom du répertoire de tri où sont rangées les images/vidéo importer",
                      action='store',
                      default='TRIAGE',
                      config="""
- value = 'directory': Sets the TRIAGE directory

  This directory is read-only
  
  This directory CAN be erased by user once processing is finished.

- value = None: no TRIAGE directory to process
  
  When verify-album is used, triage is automatically set to None in order to avoid confusion 
  between the configurations of triage and album. a Warning is issued.
  
""",
                      fr_config="""
- valeur = 'répertoire': Indique le répertoire d'où trier puis importer des images/vidéos

  Ce répertoire est en lecture seule, et PEUT être éffacé par l'utilisateur quand le piwiPre à terminé
  
- valeur = None: il n'y a pas de répertoire de tri, et donc pas d'image à importer

  Quand verify-album est utilisé, triage est automatiquement positionné à None de façon à éviter la confusion   
  entre les configurations triage et album. Un warning est généré.
""")

        self.add_item('album',
                      help='Sets the root directory for ALBUM where piwigo pictures are stored.',
                      fr_help="Indique le répertoire racine de l'ALBUM où les images sont stockées",
                      action='store',
                      default='ALBUM',
                      config="""
- value = 'directory' : Sets root directory for ALBUM
 
  a typical value is //NAS/photo, when this directory is synchronized with the piwigo repository 
  
  another typical value is ALBUM, when the piwigo repository is not accessible
  
- value =  None, the ALBUM directory is not managed, files are not copied from TRIAGE to ALBUM.""",

                      fr_config="""
- valeur = 'répertoire': Indique le répertoire où seront posées les images/vidéos

  une valeur typique est //NAS/photo, si ce répertoire est synchronisé avec le stockage de piwigo
  
  une autre valeur typique est ALBUM, si le stockage de piwigo n'est pas accessible   
  
- valeur = None: il n'y a pas de répertoire pour poser les images, et donc les images ne seront pas importées                      
""")

        self.add_item('piwigo-album-name',
                      help="Sets the name of the root piwigo album managed",
                      fr_help="Indique le nom de l'album racine de piwigo à gérer",
                      action="store",
                      default=None,
                      config="""
This item is useful only if enable-database is true.

piwiPre manages only one root piwigo album per execution (with all sub-albums),  
and MUST be coherent with the values of 'album' and 'web'

The list of first level albums can be seen through the piwigo administration web site
https://server-URL/piwigo/admin.php?page=albums

The default value is None. In this case, piwiPre will automatically select the first root album, 
which has the database global_rank "1"
""",
                      fr_config="""
Cette valeur est utile seulement lorsque enable-database est 'true'

piwiPre gère uniquement 1 album piwigo racine (avec tous ses sous-albums) par utilisation
Cette valeur DOIT être cohérente avec les valeurs de 'album' et 'web'                   
""")

        self.add_item('backup',
                      help='Sets the backup directory where modified files are saved.',
                      fr_help="Indique le répertoire de backup où sont sauvegardés le fichiers modifiés.",
                      action='store',
                      default='BACKUP',
                      config="""
- value = 'directory' : Sets the BACKUP directory, where unknown files and modified ALBUM files 
  are saved before any modification.
  
  This directory can be erased by the user once processing is finished.
  """,
                      fr_config="""
- valeur = 'répertoire' : là où sont sauvegardés les fichiers inconnus ou modifiés

  Ce répertoire PEUT être éffacé par l'utilisateur lorsque piwiPre a terminé.      
""")

        self.add_item('web',
                      help='Sets the root directory for piwigo thumbnails.',
                      fr_help="Indique le répertoire racine pour les miniatures (thumbnail) de piwigo",
                      action='store',
                      default='WEB',
                      config="""
- value = 'directory' : Sets the thumbnails directory.

  a typical value is '//NAS/web/piwigo/_data/i/galleries/photo', appropriate for synology NAS
""",
                      fr_config="""
- valeur = 'répertoire' : le répertoire racine pour les miniatures

  une valeur typique pour les NAS Synology est '//NAS/web/piwigo/_data/i/galleries/photo'      
""")

        self.add_item('remote-web',
                      help='Sets the REMOTE directory for piwigo thumbnails.',
                      fr_help="Indique le répertoire DISTANT (c-a-d sur le serveur distant) pour les miniatures",
                      action='store',
                      default=None,
                      config="""
This is useful ONLY when there is NO direct access to the piwigo thumbnails, in this case, WEB will be used
as a cache before sftp transfert.
                      
- value = 'directory' : Sets the thumbnails directory when accessed through ssh/sftp on the remote host
- if value is None, then piwigo thumbnails are NOT accessed through sftp

a typical value is '/volume1/web/piwigo/_data/i/galleries/photo', appropriate for synology NAS 
""",
                      fr_config="""
Ceci n'est utile qe lorsqu'il n'y a PAS d'accès direct au stockage des miniatures. Dans ce cas, WEB sera utilisé 
comme un cache avant le transfert via sftp.

- valeur = 'répertoire' : le répertoire DISTANT pour les miniatures
- valeur = None: il n'y a pas d'accès en sftp

Une valeur typique pour les NAS Synology est '/volume1/web/piwigo/_data/i/galleries/photo'                      
""")
        self.add_item('remote-album',
                      help='Sets the REMOTE directory for piwigo albums.',
                      fr_help="Indique le répertoire DISTANT (c-a-d sur le serveur distant) pour les photos",
                      action='store',
                      default=None,
                      )

        self.add_header("""
Management of piwigo host and users 
===================================""")

        self.add_header("""
Gestion de l'hôte piwigo et des utilisateurs 
============================================""", lang='fr')

        # --------------------------------------
        # ssh/sftp

        self.add_item('remote-user',
                      help='username on remote server, used for ssh/sftp',
                      fr_help="nom de l'utilisateur de ssh/sftp sur le serveur distant",
                      action='store',
                      default=None,
                      config="""
- Value = 'string' :username on remote server, used by ssh/sftp
- Value = None : anonymous ssh/sftp is assumed""",
                      fr_config="""
- Valeur = 'string' : utilisateur du serveur distant, pour ssh/sftp
- Valeur = None : ssh/sftp anonyme
""")

        self.add_item('remote-pwd',
                      help='Sets the password of the piwigo remote-user ',
                      fr_help="Indique le mot de passe de remote-user",
                      action='store',
                      default=None,
                      location='config')

        self.add_item('remote-host',
                      help='sets the hostname of the piwigo server, used by ssh/sftp',
                      fr_help="indique le nom du serveur ssh/sftp distant",
                      action='store',
                      default=None,
                      config="""
- Value = 'string' : hostname of the host, used by ssh
- Value = None : remote ssh cannot be used""",
                      fr_config="""
- Valeur = 'string' : nom du serveur 
- Valeur = None : ssh/sftp impossible
""")

        self.add_item('remote-port',
                      help='sets the ssh/sftp port the piwigo server',
                      fr_help="indique le port ssh du serveur distant distant",
                      action='store',
                      pwp_type=int,
                      default=42)

        self.add_item('remote-incoming',
                      help='Path, relative to the remote directory where SFTP launches, where files can be written.',
                      fr_help="Chemin, relatif au répertoire dans lequel SFTP arrive, dans lequel SFTP peut écrire",
                      action='store',
                      default=None,
                      config="""
If None, the SFTP root should be writable.

'incoming' is a typical value""",
                      fr_config="""
- Valeur = None, le repertoire racine de SFTP doit être autorisé en écriture .
- Valeur = 'incoming' est une valeur typique""")

        # --------------------------------------
        # piwigo

        self.add_item('piwigo-user',
                      help='username for piwigo access',
                      fr_help="nom d'utilisateur piwigo pour l'accès aux phots/vidéo/répertoires",
                      action='store',
                      default=None)

        self.add_item('piwigo-level',
                      help='default piwigo confidentiality level for new piwigo directories',
                      fr_help="niveau de confidentialité par défaut des nouveaux répertoires piwigo",
                      action='store',
                      default='0')
        # --------------------------------------
        # sql

        self.add_item('sql-user',
                      help='username sql server',
                      fr_help="nom de l'utilisateur du serveur SQL",
                      action='store',
                      default=None,
                      config="""
- Value = 'string' :username on sql server
- Value = None : anonymous sql access is assumed
""")

        self.add_item('sql-pwd',
                      help='Sets the password of the sql access ',
                      fr_help="mot de passe de l'utilisateur du serveur SQL",
                      action='store',
                      default=None,
                      location='config')

        self.add_item('sql-host',
                      help='sets the hostname of the sql server',
                      fr_help="nom du serveur SQL",
                      action='store',
                      default=None,
                      config="""
If None, SQL cannot be used""",
                      fr_config="""
Si None, SQL ne peut pas être utilisé""")

        self.add_item('sql-port',
                      help='sets the port for the sql server',
                      fr_help="port du serveur SQL",
                      action='store',
                      pwp_type=int,
                      default=0)

        self.add_item('sql-database',
                      help='sets the database name of the sql server',
                      fr_help="nom de la base de donnée piwigo sur le serveur SQL",
                      action='store',
                      default="piwigo")

        self.add_header("""
       
remote host configuration
=========================
Modify these settings only if you know exactly what you are doing.
The default values should be ok with any standard Linux remote host.""")

        self.add_header("""

configuration de l'hôte distant
===============================
Ne modifiez ces valeurs que si vous êtes **certain** de ce que vous faites.
Les valeurs par défaut sont correcte pour la plupart des serveurs Linux""", lang='fr')

        self.add_item('ls-command',
                      help='The remote shell command to list files.',
                      fr_help="La commande distante pour lister les fichiers",
                      action='store',
                      default='ls -sQ1HL --full-time "{file}"',
                      location='config')
        # -H    follow symbolic links listed on the command line
        # -s    print the allocated size of each file, in blocks
        # -Q    enclose entry names in double quotes
        # -1    list one file per line.  Avoid '\n' with -q or -b
        # -L    show information for the file the link references rather than for the link itself
        # --full-time  uses iso time format
        # examples:
        # '100 -rwxrwxrwx 1 foo root   98506 2024-01-17 13:28:54.779010748 +0100 "top.html"'        # noqa
        # '  4 drwxrwxr-x  2 fabien other  4096 2023-09-06 13:30:47.775207946 +0200 "Public"'       # noqa

        self.add_item('ls-output',
                      help='The output of ls-command.',
                      fr_help="Le format de sortie de ls-command",
                      action='store',
                      default=r' *\d+ +{flags} +\d+ +\w+ +\w+ +' +
                              r'{size}\s+{Y}-{m}-{d} {H}:{M}:{S}\.{ms} {z}\s+"{file}"',
                      location='config',
                      config=r"""
Where flags are taken from 
https://docs.python.org/3/library/datetime.html?highlight=datetime#strftime-strptime-behavior ,

- {dir} is 'd' for a directory
- {size} is the file size in K Bytes
- {file} is the file name
- {Y} is the year, with 4 digits
- {m} is the month number, with 2 digits
- {d} is the day number with 2 digits
- {H} is the hour, with 2 digits
- {M} the minutes, with 2 digits
- {S} the seconds, with 2 digits
- {ms} the milliseconds
- {z} the timezone, expressed as the number of hours and minutes of difference with UTC time, 
  eg. +0100 for CET during winter.
- {am} is AM or PM
- {flags} is the file mode 

Alternative for ms-dos, see https://www.windows-commandline.com/get-file-modified-date-time/

- 'dir {file}'
- '{Y}/{m}/{d} {H}:{M} {am}'\d*\s+{file}'""",
                      fr_config=r"""
Les valeurs sont similaires à: 
https://docs.python.org/3/library/datetime.html?highlight=datetime#strftime-strptime-behavior ,

- {dir} est l'indicateur 'd' de répertoire 
- {size} est la taille du fichier en K Octets
- {file} est le nom de fichier
- {Y} est l'année, sur 4 chiffres
- {m} est le numéro de mois, sur 2 chiffres
- {d} est le numéro de jour sur 2 chiffres
- {H} est l'heure, sur 2 chiffres
- {M} est la minute, sur 2 chiffres
- {S} est la seconde, sur 2 chiffres
- {ms} est la milliseconde
- {z} est le fuseau horaire, exprimé comme la difference en heures/minutes par rapport au temps UTC 
  eg. +0100 pour CET en hiver.
- {am} est AM oo PM
- {flags} est le 'mode' du fichier 

Alternative pour ms-dos, voir https://www.windows-commandline.com/get-file-modified-date-time/

- 'dir {file}'
- '{Y}/{m}/{d} {H}:{M} {am}'\d*\s+{file}'""")

        self.add_header("""
Management of actions on pictures
=================================

enable-XXX flags have 2 values:

- **false**: no action
- **true** : action is enabled if triage or album mode

By default, all actions are enabled, and this is typically done in the configuration files.


The default values enable a regular processing of ALBUM, provided **verify-album** is not empty.""")

        self.add_header("""
Gestion des actions sur les photos
==================================

les drapeaux enable-XXX ont 2 valeurs possibles:

- **false**: pas d'action
- **true** : l'action est autorisée 

Par défaut, toutes les actions sont autorisées, et ceci est changé généralement dans le fichier de configuration
dans HOME

Ces valeur par défaut sont compatibles avec les utilisations typiques,

- mode triage: trie renomme les photos/video et les envoies dans album
- mode vérification: l'album spécifié par **verify-album** est vérifié.""", lang='fr')

        self.add_item('enable-rename',
                      help='Enables files renaming',
                      fr_help="Autorise le renommage des fichiers",
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""

In album mode, pictures will **not** be moved from a directory to another, only the filename is changed""",
                      fr_config="""

En mode album, les fichiers ne sont **pas** changés de répertoire, uniquement le nom de fichier est modifié
""")

        self.add_item('enable-rotation',
                      help='Enables picture rotation',
                      fr_help="Autorise la rotation des photos",
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
                      
when ALBUM is moved from Synology photostation to piwigo, since piwigo assumes that pictures are not rotated,
enable-rotation should be used at least once per directory if not done when importing pictures.""",

                      fr_config="""

Dans le cas d'une transition entre Photostation (Synology) et piwigo, comme piwigo suppose que les 
photos ont été tournées correctement, il est prudent d'utiliser enable-rotation au moins une fois par répertoire,
de façon à assurer que les photos sont affichées correctement.
""")

        self.add_item('enable-metadata',
                      help='Enables the generation of metadata in pictures.',
                      fr_help="Autorise la génération de métadata",
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
CAVEAT, if true, the behavior can be modified by enable-metadata-reset and trust-date-in-filename""")  # FIXME ================= Start update

        self.add_item('enable-conversion',
                      help='converts pictures to JPG and video to MP4.',
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
        CAVEAT !!!
        Setting enable-conversion = false SHOULD BE AVOIDED, and used with EXTREME CARE, at your own risks.
        All potential cases HAVE NOT been tested.
        Keeping images/video formats not supported by piwiGo is NOT SAFE,
        while conversion to JPG and MP4 is straightforward and SHOULD be preferred.  
        """,
                      fr_config="""
        ATTENTION !!!
        Positionner enable-conversion = false DOIT ÊTRE ÉVITÉ, et utilisé avec une EXTRÈME PRUDENCE, 
        à vos risques et perils: Tous les cas potentiels n'ont PAS été testés.
        Garder des formats images/vidéo qui ne sont pas supportés par piwiGo N'EST PAS SÛR,  
        alors que la conversion vers JPG et MP4 est banale et DOIT être préférée.
        """)

        self.add_item('enable-thumbnails',
                      help='Enables generation of Piwigo thumbnails',
                      action='store',
                      choices=['true', 'false'],
                      default='true')

        self.add_item('enable-remote-web',
                      help='Enables the copy of piwigo thumbnails with ssh/sftp.',
                      action='store',
                      choices=['true', 'false'],
                      default='false')

        self.add_item('enable-remote-album',
                      help='Enables the copy of piwigo pictures/video with ssh/sftp.',
                      action='store',
                      choices=['true', 'false'],
                      default='false')

        self.add_item('enable-database',
                      help='Enables the verification of database information.',
                      action='store',
                      choices=['true', 'false'],
                      default='false')

        self.add_item('enable-auto-configuration',
                      help='Enables configuration of ALBUM from TRIAGE, by creating a configuration file in WEB.',
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
Enables the copy of piwiPre.ini files found in TRIAGE directory to the corresponding folder of WEB,
so that further processing of ALBUM give the same results.""")

        self.add_item('trust-date-in-filename',
                      help="if enable-metadata is true, reads the picture date in the filename (vs in the metadata)",
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
CAVEAT: This flag should be set to false WITH CARE!
 
If there is a date in the filename, (according to the 'names' argument), then this date is used for metadata
Else, if there is a date in metadata, this one is kept
Else, the file creation time is used, and written in metadata
So, if a file is checked twice, the 2nd run does not perform any change

Use it ONLY when the metadata is known to be wrong, and the filename has been manually set.
It is a good practice to store it in the auto-config piwiPre.ini file
""")

        self.add_item('enable-metadata-reset',
                      help="if enable-metadata is true, then metadata can be overwritten",
                      action='store',
                      choices=['true', 'false'],
                      default='false',
                      config="""
This flag is used only when enable-metadata is true.
    - value = false (the default): metadata is written in the file ONLY if there was no metadata
    - value = true: metadata is written if different from what was already in the file 
      (which includes no value)""")

        self.add_item('enable-pwg-representative',
                      help="enables the creation of piwigo JPG representative of video",
                      fr_help="autorise la creation de l'image JP qui représente une vidéo dans piwigo",
                      action='store',
                      choices=['true', 'false'],
                      default='true')

        self.add_item('ffmpeg-path',
                      help="path to ffmpeg executable, should end with a / if not empty",
                      action='store',
                      default=(os.environ['PROGRAMFILES(X86)'] + '/ffmpeg/bin/') if platform.system() == "Windows"  # noqa
                      else '/usr/bin/',
                      config="""
- ffmpeg and ffprobe are used to handle video files.
- The default path should be OK for your system (windows or linux)
- If ffmpeg and ffprobe are in the PATH, you can leave ffmpeg-path empty
""")
        self.add_item('git-path',
                      help="path to git executable on WINDOWS, for INTERNAL use only",
                      action='store',
                      default=None,
                      config="""
This flag is used internally ONLY with --install-exe, because piwiPre is called in the administrator context
Under normal circumstances, you should NOT use it 
""")

        self.add_header("""
configuration only
==================

The following configuration items are not accessible through command line options
and must be specified in a configuration file.""")

        self.add_item('names',
                      help='The format of renamed pictures. This includes the path starting from ALBUM.',
                      action='store',
                      default='{Y}/{Y}-{m}-{month_name}-{d}-{base}/{Y}-{m}-{d}-{H}h{M}-{S}-{base}.{suffix}',
                      location='config',
                      config=r"""

CAVEAT: The value must be enclosed in single or double quotes !                      
                      
Field values:

- {Y} etc are inherited from the IPTC date of the picture.

- {base} is the name of the TRIAGE folder where the picture was originally found.

- {author} is computed according to the camera name in the IPTC metadata, see **authors**

- {count} is the current count of pictures in the directory,
  so that it is 01 for the first picture, 02 for the 2nd etc.

- {suffix}: file suffix, typically jpg, txt, mp4...

- All numeric fields are printed with 2 digits, excepted year which has 4.

When several  different pictures are supposed to have the same filename,
the last numeric field (here {s}) is incremented until a unique filename is found.


Many users prefer names that include the date,
so that name collisions are avoided when pictures are out in a flat folder.

But different schemes are possible.
For instance, "{Y}/{m}/{d}/{base}-{count}", is also a traditional naming.

all characters that are not in 
"a-zA-Z0-9\-_.&@~!,;+°()àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ " will be replaced by '_' 
""")  # noqa

        self.add_item('month-name',
                      help='The name for each month, used to rename files.',
                      fr_help='Le nom de chaque mois, utilisé renommer les fichiers.',
                      action='store',
                      pwp_type=list,
                      default=['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
                      fr_default = ['Janvier',  'Février',  'Mars',  'Avril', 'Mai', 'Juin', 'Juillet',       # noqa
                                    'Août', 'Septembre', 'Octobre', 'Novembre', 'Décembre'],
                      location='config')

        self.add_item('authors',
                      help='A dictionary of mappings between camera model name as found in Exif data, and author name',
                      action='store',
                      pwp_type=dict,
                      default={},
                      location='config',
                      config="""
- example of possible value ::

   Camera1 : author1
   'Camera 2' : 'author 2'
   DEFAULT : 'default value'""")

        self.add_item('copyright',
                      help='A copyright sentence to be written in Exif metadata, with obvious fields.',
                      action='store',
                      default="(C) {author} {Y}",
                      location='config',
                      config="""- The date is taken from the photo metadata, {month} and {day} are also available.'""")

        self.add_item('instructions',
                      help="A sentence to be written in Exif metadata, with {author} coming from the 'authors' section",
                      action='store',
                      default="No copy allowed unless explicitly approved by {author}",
                      fr_default="Aucune copie autorisée sauf si explicitement approuvée par {author}",
                      location='config',
                      config="""- adding an email or a phone number may be appropriate.""")

        self.add_item('dates',
                      help='A dictionary of dates corrections',
                      action='store',
                      pwp_type=dict,
                      default='',  # {},
                      location='config',
                      config="""
Date corrections are used only to compute the new name of the picture in the renaming step.
The metadata (notably dates) of the picture is unchanged.

- each toplevel item a dictionary with a unique name
- each date is written as a dictionary with year, month, day, hour, minute, second, some items may be missing

- the dictionary describes each correction with the following fields:

  - 'start', 'end': dates. the correction occurs if the picture date is between these boundaries
  - camera_name: the name of the camera for this correction or 'default' for camera name not in the list
  - 'delta' or 'forced' : a date. 
  
    - If 'delta', the date is added to the picture date. 
    - If  'forced' the picture date is set to this value.

- the specific 'NO-DATE' toplevel item is for pictures without a date.
   - the 'start', 'end', delta dates are not defined
   - this item contains only 'forced' date that will be set to all pictures without a date
- when a date is 'forced', and hour, minute, second are not specified, piwiPre uses the picture creation time.
   
See also the online documentation 
   """)

        self.add_header("""
example ::

    dates:
        USA:                 # this name should be unique within the 'dates'
            start:
                year:  2018
                month:  7
                day: 4
                hour: 20
            end:
                year:  2018
                month:  7
                day: 6
                hour: 23
            D6503:              # camera name
                delta:
                    hour: 9
            TG-320:            # a different camera
                delta:
                    hour: 9
                    minute: 30
        Utah 1:
            start:
                year:  2018
                month:  7
                day: 6
                hour: 23
            end:
                year:  2018
                month:  7
                day: 8
                hour: 23
            TG-320 :
                delta:
                    hours: 8   # CAVEAT: here, hours and not hour ! (and years, etc...)
        NO-DATE:               # CAVEAT: like python, yaml is strict on indentation errors
            forced :
                 year: 2023
                 month: 7
                 day : 24

.. Note:: usually, 'NO-DATE' and  'forced' are not set on a global ALBUM base, 
   but rather in a specific TRIAGE or ALBUM folder where abnormal pictures are known to be stored.
   

.. Important:: unless enable-auto-configuration == false,  
   when a .ini file is stored in a TRIAGE folder or sub-folder, 
   then it  will be copied in the corresponding WEB subdirectories,  
   so that further processing of ALBUM give the same results. 
   This is particularly useful for dates management
""")

        self.add_item('piwigo-thumbnails',
                      help="A dictionary of piwigo thumbnails to be built, including formats",
                      action='store',
                      pwp_type=dict,
                      default={
                          "{f}-sq.jpg": {'width': 120, 'height': 120, 'crop': True},
                          "{f}-th.jpg": {'width': 144, 'height': 144, 'crop': False},
                          "{f}-me.jpg": {'width': 792, 'height': 594, 'crop': False},
                          "{f}-cu_e250.jpg": {'width': 250, 'height': 250, 'crop': True},
                      },
                      location='config',
                      config="""
A dictionary if thumbnail specifications,
- {f} is the photo basename
- width = maximum width
- height = maximum height
- crop = the picture will be cropped to a square form factor.

The regular piwigo thumbnails defined in the documentation are as follows ::

    "{f}-sq.jpg" : 120, 120, crop      # SQUARE: mandatory format
    "{f}-th.jpg":  144, 144            # THUMB:  mandatory
    "{f}-me.jpg" : 792, 594            # MEDIUM: mandatory
    "{f}-2s.jpg" : 240, 240            # XXSMALL       # noqa
    "{f}-xs.jpg" : 432, 324            # XSMALL        # noqa
    "{f}-sm.jpg" : 576, 432            # SMALL
    "{f}-la.jpg" : 1008, 756           # LARGE
    "{f}-xl.jpg" : 1224, 918           # XLARGE        # noqa
    "{f}-xx.jpg" : 1656, 1242          # XXLARGE       # noqa
    "{f}-cu_e250.jpg" : 250, 250, crop # CU    : mandatory"""),  # noqa

        self.add_header("""
        Language management and Misc
        ============================
        """)
        self.add_item('language',
                      help="sets the language for help and a few options",
                      fr_help="change la langue pour l'aide et quelques options",
                      action='store',
                      choices=['en', 'fr'],
                      default="en",
                      fr_default='fr',
                      config="""
Changing the language as an effect on the **default** values of **names**, **month-name**, **copyright**
and  **--help** prints the help in the chosen language
""",
                      fr_config="""
Changer la langue a un effet sur les valeurs **par défaut** des options suivantes: **names**, **month-name**, 
**copyright** et  **--help** imprime l'aide dans la langue choisie.

Le nom des options n'est PAS traduit, par exemple --help reste --help, ne devient pas --aide
""")
        self.add_item('enable-colors',
                      help="Prints output with colors",
                      fr_help="Imprime sur le terminal avec des couleurs",
                      action='store',
                      choices=['true', 'false'],
                      default='true')

        self.config = self.parse_args_and_ini(program, self.default_ini_file, arguments, with_config=with_config)


def build_official_rst(autotest: bool):
    if autotest:
        filename_en = "tests/results/configuration.rst"
        filename_fr = "tests/results/configuration_fr.rst"
    else:
        filename_en = 'source/usage/configuration.rst'
        filename_fr = 'source/fr/configuration.rst'

    source = 'piwiPre/pwpParser.py'
    if not autotest and os.path.getmtime(filename_en) > os.path.getmtime(source):
        LOGGER.msg(f"file '{filename_en}' is older than source '{source}': patch useless")
        return
    parser = PwpParser(arguments=[], with_config=True, program="autotest")
    LOGGER.msg(f"building english rst '{filename_en}' from '{source}'")
    parser.build_rst(filename_en, lang='en')
    LOGGER.msg(f"construction du rst français '{filename_en}' depuis '{source}'")
    parser.build_rst(filename_fr, lang='fr')


def pwp_parser_main(arguments):
    LOGGER.msg('--------------- starting pwp_test_config')
    parser = PwpParser(arguments=arguments, program="parser_autotest", with_config=False)
    config = parser.parse_args_and_ini("test harness", "tests.ini", arguments)
    rst = "../results/test-result.rst"
    ini = "../results/test-result.ini"
    parser.build_rst(rst)
    parser.build_ini_file(ini)
    pprint.pprint(config)
    parser.print_help()
    LOGGER.msg('--------------- end of  pwp_test_config')


if __name__ == "__main__":
    sys.exit(pwp_parser_main(sys.argv[1:]))
